//---------------------------------------------
// FracNEON_dp_opt3 Benchmark by Michael Kübel 
//---------------------------------------------

#include <iostream>
#include <iomanip>
#include <cstdint>
#include <SDL2/SDL.h>
#define WINDOW_WIDTH 600

using namespace std;

extern "C" uint64_t CalcMandelbrotASM_FracNEON_dp_opt3(const uint32_t* p);

int main(int argc, char** argv)
{
	const char prg_info[] = "FracNEON_dp_opt3 V1.0 by Michael Kübel";
    uint32_t *pixels = new uint32_t[WINDOW_WIDTH*WINDOW_WIDTH];  //32 Bit pixel array mapped to texture
    memset(pixels,0,WINDOW_WIDTH*WINDOW_WIDTH*sizeof(uint32_t)); //set all to pixels to 0
	uint64_t global_iterations = {0};                            //64 bit global counter address
    unsigned int repeat_benchmark;

    SDL_Event event;
    SDL_Init(SDL_INIT_VIDEO);
    SDL_Window   *window   = SDL_CreateWindow(prg_info,SDL_WINDOWPOS_UNDEFINED,SDL_WINDOWPOS_UNDEFINED,WINDOW_WIDTH,WINDOW_WIDTH,0);
    SDL_Renderer *renderer = SDL_CreateRenderer(window,-1,0);
    SDL_Texture  *texture  = SDL_CreateTexture(renderer,SDL_PIXELFORMAT_ARGB8888,SDL_TEXTUREACCESS_STATIC,WINDOW_WIDTH,WINDOW_WIDTH);
    SDL_SetRenderDrawColor(renderer,0,0,0,0);
    SDL_RenderClear(renderer);
    SDL_RenderPresent(renderer);

    cout<<prg_info<<"\n";
    cout<<"--------------------------------------\n";
    cout<<"Calculating a Mandelbrot of 600x600 pixels at 4096 iterations with double-precision floats\n";

	cout<<"Calibrate timing...\n";
	uint64_t calibrate_start = SDL_GetPerformanceCounter();
	for (unsigned int i=0;i<10;i++) {
		global_iterations = CalcMandelbrotASM_FracNEON_dp_opt3(pixels);}
	uint64_t calibrate_end = SDL_GetPerformanceCounter();
	cout<<"Done...\n";
	
	float calibration_timing =(calibrate_end-calibrate_start)/(float)SDL_GetPerformanceFrequency();
	repeat_benchmark=static_cast<int>(100/calibration_timing);
	if (repeat_benchmark<1) {repeat_benchmark=1;}

	//Start benchmark with calculated repeats to reach about 10 seconds of benchmark timing
    cout<<"Benchmarking for around 10 seconds while looping "<<repeat_benchmark<<" times...\n";
    global_iterations=0;  // reset global iterations
	uint64_t benchmark_start = SDL_GetPerformanceCounter();
    for (unsigned int f=0;f<repeat_benchmark;f++) {
		global_iterations += CalcMandelbrotASM_FracNEON_dp_opt3(pixels);} // add iterations
    uint64_t benchmark_end = SDL_GetPerformanceCounter();
    
    cout<<"Done...\n";

    float benchmark_timing =(benchmark_end-benchmark_start)/(float)SDL_GetPerformanceFrequency();
    SDL_UpdateTexture(texture,NULL,pixels,WINDOW_WIDTH*sizeof(uint32_t)); //NULL = entire texture
    SDL_RenderCopy(renderer,texture,NULL,NULL); //NULL = entire texture, NULL = entire destination
    SDL_RenderPresent(renderer);
    cout<<"Iterations............: "<<global_iterations<<"\n";
    cout<<"Time taken............: "<<fixed<<setprecision(3)<<benchmark_timing<<" s\n";
    cout<<"Speed absolute........: "<<fixed<<setprecision(3)<<(global_iterations/benchmark_timing/1000000)<<" Million Iterations/s\n";
    cout<<"Close graphics window to quit...\n";
    while (1) {if (SDL_PollEvent(&event) && event.type == SDL_QUIT) break;}

    delete[] pixels;
    SDL_DestroyTexture(texture);
    SDL_DestroyRenderer(renderer);
    SDL_DestroyWindow(window);
    SDL_Quit();

    return 0;
}
