//---------------------------------------------------------
// FracNEON_dp_opt2 *text only* Benchmark by Michael Kübel 
//---------------------------------------------------------

#include <iostream>
#include <iomanip>
#include <cstdint>
#include <chrono>
#include <cstring>
#define WINDOW_WIDTH 600

using namespace std;

extern "C" uint64_t CalcMandelbrotASM_FracNEON_dp_opt2(const uint32_t* p);

int main(int argc, char** argv)
{
    uint32_t *pixels = new uint32_t[WINDOW_WIDTH*WINDOW_WIDTH];
    memset(pixels,0,WINDOW_WIDTH*WINDOW_WIDTH*sizeof(uint32_t));
	uint64_t global_iterations = {0};                            //64 bit global counter address
    unsigned int repeat_benchmark;

    cout<<"FracNEON_dp_opt2 V1.0 by Michael Kübel\n";
    cout<<"--------------------------------------\n";
    cout<<"Calculating a Mandelbrot of 600x600 pixels at 4096 iterations with double-precision floats\n";
	cout<<"Calibrate timing...\n";
	auto start1 = std::chrono::system_clock::now();
	for (unsigned int i=0;i<10;i++) {
		global_iterations = CalcMandelbrotASM_FracNEON_dp_opt2(pixels);}
    auto end1 = std::chrono::system_clock::now();
    chrono::duration <double> calibration_timing = end1-start1;	
    cout<<"Done...\n";

    repeat_benchmark=static_cast<int>(100/calibration_timing.count());
	if (repeat_benchmark<1) {repeat_benchmark=1;}

    cout<<"Benchmarking for around 10 seconds while looping "<<repeat_benchmark<<" times...\n";
    global_iterations=0; // reset global iterations
 	auto start2 = std::chrono::system_clock::now();
    for (unsigned int f=0;f<repeat_benchmark;f++) {
		global_iterations += CalcMandelbrotASM_FracNEON_dp_opt2(pixels);}
    auto end2 = std::chrono::system_clock::now();
    chrono::duration <double> benchmark_timing = end2-start2;	
    cout<<"Done...\n";

    cout<<"Iterations............: "<<global_iterations<<"\n";
    cout<<"Time taken............: "<<fixed<<setprecision(3)<<benchmark_timing.count()<<" s\n";
    cout<<"Speed absolute........: "<<fixed<<setprecision(3)<<(global_iterations/benchmark_timing.count()/1000000)<<" Million Iterations/s\n";
    delete[] pixels;
    return 0;
}
