//-------------------------------------------------
// FracNEON_dp_v82_mt1 Assembler routine
//-------------------------------------------------

armv8_2        = true // ARMv8.2 compativle CPU ?  
fractal_size   =  600 // size of fractal in pixels
max_iterations = 4096 // maximum iterations
colour_shift   =    3 // shift colours left

.text

.global CalcMandelbrotASM_FracNEON_dp_v82_mt1
CalcMandelbrotASM_FracNEON_dp_v82_mt1:

// to allow use of all regs and call further subroutines,
// backup x17...x30, d8...d15 due to calling conventions
// (x17 for align reasons and x18 due to system variations)
 STP x17,x18,[sp,#-224]!
 STP x19,x20,[sp,#  16]
 STP x21,x22,[sp,#  32]
 STP x23,x24,[sp,#  48]
 STP x25,x26,[sp,#  64]
 STP x27,x28,[sp,#  80]
 STP x29,x30,[sp,#  96]
 STP  d8, d9,[sp,# 112]
 STP d10,d11,[sp,# 128]
 STP d12,d13,[sp,# 144]
 STP d14,d15,[sp,# 160]
 STP x1,x2,  [sp,# 176] // X1 global address for iteration counter
                        // X2 global address for y 

 MOV X26,X0             // use X26 for pixel_buffer start address as X0 is used elsewhere
 MOV X13,X2             // backup address of global iteration counter as X2 is used elsewhere  
 MOV X27,X3             // set first y in X5 as X3 is used elsewhere
 MOV X4,#0              // init local iteration counter
 LDR X28,=frac_data     // pointer to fractal data
      
   //*** D20 = 'ru' will be global constant *** 
   LDP D0,D20,[X28]      // ro (D0), ru (D20)
   LDR D2,[X28,#24]      // fractal size
   FSUB D0, D0,D20
   //*** D21 = 'delta_sp' will be global constant ***
   FDIV D21,D0, D2       // delta_sp = (ro-ru)/fractal_size
   FMOV v15.2d,#4.0      // limit = 4.0 for each pixel

   MOVI v14.2s,#4               // init sub value = 4 for each pixel (4 times loop unrolling)   
   MOV W12,#(max_iterations-1)  // init maximum iterations - 1

   y_loop:

      MOV W6,#fractal_size  // get pixel_buffer_width
      MUL W6,W6,W27         // pixel_buffer_width * y_counter 
      ADD X25,X26,W6,UXTW#2 // Line position = pixel_buffer + (pixel_buffer_width * y)<<2
   
      LDR D0,[X28,#16]     // io

      UCVTF D2,W27         // get y  
      FMUL  D2, D2,D21     // use FMSUB later	  
      //*** S22 = 'io_save_sp' will be global constant *** 
      FSUB  D22,D0, D2     // b0 = io - current_y * delta
      DUP v5.2d,v22.d[0]   // save io_save_sp = b0=b1 in Q5

      MOV W6 ,#0           // starting line counter pixel 1
      MOV W7, #1           // starting line counter pixel 2
      MOV W10,W7           // init line counter global
  
      LDR  q1,[X28,#32]         // v1 =  0.0      1.0
      DUP  v0.2d,v20.d[0]       // v0 =   ru       ru
      FMUL v1.2d,v1.2d,v21.d[0] // v1 =  d*x      d*(x+1)
      FADD v0.2d,v0.2d,v1.2d    // v0 = ru+d*x  ru+d*(x+1)
                                // x  = a ...*** use FMLA later...
	  
      MOV v4.16B,v0.16B      // save a0,a1
      MOV v1.16B,v5.16B      // y = b0

      DUP v2.2s,W12          // init single iteration counters = max_iter-1 ***use vector, upper 64 Bit will be zero

      MOV W11,#0             // init end of line flags for each pixel

      iteration_loop:
         //1.Iteration
         FMUL  v3.2d, v0.2d, v0.2d   // x*x
         FMUL  v8.2d, v1.2d, v1.2d   // y*y
         FMUL v12.2d, v0.2d, v1.2d   // x*y
         FADD v10.2d, v3.2d, v8.2d   // x*x + y*y
         FSUB  v9.2d, v3.2d, v8.2d   // x_new = x*x - y*y
         FADD v12.2d,v12.2d,v12.2d   // 2*x*y
         FADD  v0.2d, v9.2d, v4.2d   // x_new = x_new + a
         FADD  v1.2d,v12.2d, v5.2d   // y_new = y_new + b

         //2.Iteration
         FMUL  v3.2d, v0.2d, v0.2d   // x*x
         FMUL  v8.2d, v1.2d, v1.2d   // y*y
         FMUL v12.2d, v0.2d, v1.2d   // x*y
         FADD v11.2d, v3.2d, v8.2d   // x*x + y*y
         FSUB  v9.2d, v3.2d, v8.2d   // x_new = x*x - y*y
         FADD v12.2d,v12.2d,v12.2d   // 2*x*y
         FADD  v0.2d, v9.2d, v4.2d   // x_new = x_new + a
         FADD  v1.2d,v12.2d, v5.2d   // y_new = y_new + b

         //3.Iteration
         FMUL  v3.2d, v0.2d, v0.2d   // x*x
         FMUL  v8.2d, v1.2d, v1.2d   // y*y
         FMUL v12.2d, v0.2d, v1.2d   // x*y
         FADD  v6.2d, v3.2d, v8.2d   // x*x + y*y
         FSUB  v9.2d, v3.2d, v8.2d   // x_new = x*x - y*y
         FADD v12.2d,v12.2d,v12.2d   // 2*x*y
         FADD  v0.2d, v9.2d, v4.2d   // x_new = x_new + a
         FADD  v1.2d,v12.2d, v5.2d   // y_new = y_new + b

         //4.Iteration
         FMUL  v3.2d, v0.2d, v0.2d   // x*x
         FMUL  v8.2d, v1.2d, v1.2d   // y*y
         FMUL v12.2d, v0.2d, v1.2d   // x*y
         FADD  v7.2d, v3.2d, v8.2d   // x*x + y*y
         FSUB  v9.2d, v3.2d, v8.2d   // x_new = x*x - y*y
         FADD v12.2d,v12.2d,v12.2d   // 2*x*y
         FCMGT v3.2d, v7.2d,v15.2d   // > 4 ?
         FADD  v0.2d, v9.2d, v4.2d   // x_new = x_new + a
         SUB   v2.2s, v2.2s,v14.2s   // decrement pixel counters
         FADD  v1.2d,v12.2d, v5.2d   // y_new = y_new + b

         UMAXV s17,v3.4s             // works also for double
         UMAXV s16,v2.4s             // works also for double       
         MOV   w17,v17.s[0]
         MOV   w16,v16.s[0]
 
         TST w17,w17            // faster than TBNZ here
         B.NE  diverged_12      // 1 or more pixel diverged
         TST w16,w16            // faster than TBNZ here
      B.PL iteration_loop
      B max_iterations_12    // 1 or more pixel reached max iter
 
      //############ Pixel 1 Diverge Handling ############

      diverged_12:
      FCMP D7,D15                // > 4 ? pixel 1 diverged   v7.2d[0]
      B.LE diverged_2            // not ? -> check pixel 2,3,4
  
      UMOV  W0,v2.s[0]         // get iterations pixel 1   v2.2d[0]	 
      FCMP  D6,D15             // > 4 ? Diverged at 3rd ?  v6.2d[0]
      CSINC W0,W0,W0,LE        // if => inc counter
      FCMP  D11,D15            // > 4 ? Diverged at 2nd ? v11.2d[0]
      CSINC W0,W0,W0,LE        // if => inc counter
      FCMP  D10,D15            // > 4 ? Diverged at 1st ? v10.2d[0]
      CSINC W0,W0,W0,LE        // if => inc counter

        SUB   W2,W12,W0          // correct counter
        ADD   W10,W10,#1         // inc x
        ADD   W4,W4, W2          // add to global iteration counter

        LSL   W2,W2,#colour_shift// some colouring
        STR   W2,[X25,W6,UXTW#2] // write pixel to buffer

        CMP   W10,#fractal_size  // end of line reached ?
        B.LT skip_endofline_p1
          ORR W11,W11,#0b01      // set end of line flags for pixel 1
          B   diverged_2         // but still check if also another diverged
        skip_endofline_p1:

        INS   v2.s[0],W12        // transfer counter init to v2.s[0]
        MOV   W6,W10             // save new x in new plot_x1
        UCVTF D23,W10
        FMUL  D23,D23,D21        // new x1 = x1*d
        FADD  D23,D23,D20        // new x1 = ru + x1*d
        INS   v0.d[0],v23.d[0]   // place new x1 in correct vector v0.d[0]
        INS   v4.d[0],v23.d[0]   // new x1 also = new 'a'    
        INS   v1.d[0],v22.d[0]   // io_save_sp => get saved y4=y3=y2=y1
 
         //just continue checking pixel 2

      //############ Pixel 2 Diverge Handling ############

      diverged_2:
      INS  v23.d[0],v7.d[1]      // get pixel 2 from v7.4[1]
      FCMP D23,D15               // > 4 ? pixel 2 diverged
      B.LE end_diverge_check     // not ? 

      INS   v23.d[0], v6.d[1]  // place out of order
      INS   v24.d[0],v11.d[1]
      INS   v25.d[0],v10.d[1]
      UMOV  W0,v2.s[1]         // get iterations pixel 2   v2.sd[1]	 
      FCMP  D23,D15            // > 4 ? Diverged at 3rd ?  v6.2d[1]
      CSINC W0,W0,W0,LE        // if => inc counter
      FCMP  D24,D15            // > 4 ? Diverged at 2nd ? v11.2d[1]
      CSINC W0,W0,W0,LE        // if => inc counter
      FCMP  D25,D15            // > 4 ? Diverged at 1st ? v10.2d[1]
      CSINC W0,W0,W0,LE        // if => inc counter

        SUB   W2,W12,W0          // correct counter
        ADD   W10,W10,#1         // inc x
        ADD   W4,W4, W2          // add to global iteration counter

        LSL   W2,W2,#colour_shift// some colouring
        STR   W2,[X25,W7,UXTW#2] // write pixel to buffer

        CMP   W10,#fractal_size  // end of line reached ?
        B.LT skip_endofline_p2
          ORR W11,W11,#0b10      // set end of line flags for pixel 2
          B   end_diverge_check  // but still check if also another diverged
        skip_endofline_p2:

        INS   v2.s[1],W12        // transfer counter init to v2.s[1]
        MOV   W7,W10             // save new x in new plot_x2
        UCVTF D23,W10
        FMUL  D23,D23,D21        // new x2 = x2*d
        FADD  D23,D23,D20        // new x2 = ru + x2*d
        INS   v0.d[1],v23.d[0]   // place new x1 in correct vector v0.d[1]
        INS   v4.d[1],v23.d[0]   // new x2 also = new 'a'    
        INS   v1.d[1],v22.d[0]   // io_save_sp => get saved y4=y3=y2=y1

      end_diverge_check:

      CBZ W11,iteration_loop     // was end of line reached ?

      //############ Pixel 1 Max Iterations Handling ############

      max_iterations_12: 
      MOV  W0,v2.s[0]            // check if pixel 1 reach max iter
      TBZ W0,31,max_iterations_2 // not ? -> check pixel 2

      ADD W10,W10,#1           // inc x
      ADD W4,W4,#max_iterations// add to global iteration counter
      CMP W10,#fractal_size    // end of line reached ?
      B.LT max_skip_endofline_p1
        ORR W11,W11,#0b01      // set end of line flags for pixel 1
        B   max_iterations_2
      max_skip_endofline_p1:

      MOV   W6,W10             // save new x in new plot_x1
      INS   v2.s[0],W12        // transfer counter init to v2.d[0]
      UCVTF D23,W10
      FMUL  D23,D23,D21        // new x1 = x1*d
      FADD  D23,D23,D20        // new x1 = ru + x1*d
      INS   v0.d[0],v23.d[0]   // place new x1 in correct vector v0.d[0]
      INS   v4.d[0],v23.d[0]   // new x1 also = new 'a'    
      INS   v1.d[0],v22.d[0]   // io_save_sp => get saved y4=y3=y2=y1

      //############ Pixel 2 Max Iterations Handling ############

      max_iterations_2:
      MOV  W0,v2.s[1]              // check if pixel 2 reach max iter
      TBZ W0,31,end_max_iter_check // not ? -> end_max_iter_check

      ADD W10,W10,#1           // inc x
      ADD W4,W4,#max_iterations// add to global iteration counter
      CMP W10,#fractal_size    // end of line reached ?
      B.LT max_skip_endofline_p2
        ORR W11,W11,#0b10      // set end of line flags for pixel 2
        B   end_max_iter_check
      max_skip_endofline_p2:

      MOV   W7,W10             // save new x in new plot_x2
      INS   v2.s[1],W12        // transfer counter init to v2.d[1]
      UCVTF D23,W10
      FMUL  D23,D23,D21        // new x2 = x2*d
      FADD  D23,D23,D20        // new x2 = ru + x2*d
      INS   v0.d[1],v23.d[0]   // place new x1 in correct vector v0.d[1]
      INS   v4.d[1],v23.d[0]   // new x2 also = new 'a'    
      INS   v1.d[1],v22.d[0]   // io_save_sp => get saved y4=y3=y2=y1

      end_max_iter_check:

      CBZ W11,iteration_loop

      end_of_line:

      MOV v24.16B,v0.16B // save state of v0 for other pixels
      MOV v25.16B,v1.16B // save state of v1 for other pixels
 
      //############ End of Line Pixel 1 Handling  ############
      
      TBNZ W11,#0,check_end_of_line_pixel_2 //test for 0b0001

      MOV W1,v2.s[0]  
      iterate_pixel_1:
         FMUL  D3, D0, D0 // x*x
         FMUL  D8, D1, D1 // y*y
         SUB   W1, W1, #1 // decrement pixel counter
         FMUL D12, D0, D1 // x*y
         FADD  D7, D3, D8 // x*x + y*y
         FSUB  D9, D3, D8 // x_new = x*x - y*y
         FADD D12,D12,D12 // 2*x*y
         FADD  D0, D9, D4 // x_new = x_new + a
         FCMP  D7,D15     // pixel 1 > 4 ? check if pixel 1 diverged
         FADD  D1,D12, D5 // y_new = y_new + b
         B.GT  end_pixel_1_diverged
         TBZ W1,31,iterate_pixel_1

	     //maximum iterations reached
         ADD W4,W4,#max_iterations   // add to global iteration counter
         B check_end_of_line_pixel_2

         end_pixel_1_diverged:
         SUB W2,W12,W1          // correct counter
         ADD W4,W4, W2          // add to global iteration counter
         LSL W2,W2,#colour_shift// some colouring
         STR W2,[X25,W6,UXTW#2] // write pixel to buffer

      //############ End of Line Pixel 2 Handling  ############

      check_end_of_line_pixel_2:

      MOV v0.16B,v24.16B // restore state of v0 for other pixels
      MOV v1.16B,v25.16B // restore state of v1 for other pixels
 	  
      TBNZ W11,#1,next_line

      INS v0.d[0],v0.d[1] 
      INS v1.d[0],v1.d[1] 
      INS v4.d[0],v4.d[1] 
      MOV W1,v2.s[1]  
      iterate_pixel_2:
         FMUL  D3, D0, D0 // x*x
         FMUL  D8, D1, D1 // y*y
         SUB   W1, W1, #1 // decrement pixel counter
         FMUL D12, D0, D1 // x*y
         FADD  D7, D3, D8 // x*x + y*y
         FSUB  D9, D3, D8 // x_new = x*x - y*y
         FADD D12,D12,D12 // 2*x*y
         FADD  D0, D9, D4 // x_new = x_new + a
         FCMP  D7,D15     // pixel 1 > 4 ? check if pixel 1 diverged
         FADD  D1,D12, D5 // y_new = y_new + b
         B.GT  end_pixel_2_diverged
         TBZ W1,31,iterate_pixel_2
 
         // maximum iterations reached
         ADD W4,W4,#max_iterations   // add to global iteration counter
         B next_line

         end_pixel_2_diverged:
         SUB W2,W12,W1          // correct counter
         ADD W4,W4, W2          // add to global iteration counter
         LSL W2,W2,#colour_shift// some colouring
         STR W2,[X25,W7,UXTW#2] // write pixel to buffer

   next_line:

   // atomic update of global y_line counter
.if armv8_2 == true
   MOV     W6,#1
   LDADDAL W6,W27,[X13]   //[X13], [X13] = [X13] + 1
   ADD     W27,W27,#1     //W5=W5+1 for CMP check below
.else
   try_again_update_global_y_line_counter:
     LDAXR W6,[X13]        //try LDXR or LDAXR
     ADD   W27,W6,#1
     STLXR W7,W27,[X13]     //try STXR or STLXR
   CBNZ W7,try_again_update_global_y_line_counter
.endif
   CMP W27,#fractal_size   //may be check can already be done within the atomic update, check performance
   B.LO y_loop             //use lower check due to overflow of other cores
   
// prepare to return to C routine
// restore x17...x30, d8...d15
   LDP x17,x18,[sp],#16
   LDP x19,x20,[sp],#16
   LDP x21,x22,[sp],#16
   LDP x23,x24,[sp],#16
   LDP x25,x26,[sp],#16
   LDP x27,x28,[sp],#16
   LDP x29,x30,[sp],#16
   LDP  d8, d9,[sp],#16
   LDP d10,d11,[sp],#16
   LDP d12,d13,[sp],#16
   LDP d14,d15,[sp],#16
   LDP x1, x2, [sp],#48  //get address for iter count storage

// atomic update of global iteration counter
.if armv8_2 == true
   LDADDAL X4,XZR,[X1]     //XZR(not needed) = [X1], [X1] = [X1] + X4
.else
   try_again_update_global_iteration_counter:
     LDAXR X9,[X1]         //try LDXR or LDAXR
     ADD   X6,X9,X4
     STLXR W7,X6,[X1]      //try STXR or STLXR
   CBNZ W7,try_again_update_global_iteration_counter
.endif
RET

.data
.align 4
frac_data:
.double -0.1450 //+ 0 ro
.double -0.1750 //+ 8 ru
.double -1.0200 //+16 io
.double   600.0 //+24 fractal_size
.double    0.0  //+32 
.double    1.0
