//-------------------------------------------------
// FracNEON_dp_v82_mt2 Assembler routine
//-------------------------------------------------
//
armv8_2        = true // ARMv8.2 compativle CPU ?  
fractal_size   =  600 // size of fractal in pixels
max_iterations = 4096 // maximum iterations
colour_shift   =    3 // shift colours left
//  v0 = x01
//  v1 = y01
//  v2 = calc p01 (x*x),...
//  v3 = calc p01 (y*y),...
//  v4 = calc p01 (x*x),...
//  v5 = x*x+y*y save iter p01 0
//  v6 = x*x+y*y save iter p01 1
//  v7 = counter p01
//  v8 = a01
//
//  v9 = x23
// v10 = y23
// v11 = calc p23 (x*x),...
// v12 = calc p23 (y*y),...
// v13 = calc p23 (x*x),...
// v14 = x*x+y*y save iter p23 0
// v15 = x*x+y*y save iter p23 1
// v16 = counter p23
// v17 = a23
//
// v27 = b0123
//                 [1]|          [0]
// v28 =           4.0|         4.0  diverge limit float
//          [3]|   [2]|   [1]|   [0]
// v29 =      -|     -|     4|     4   counter add integer
//                 [1]|          [0]
// v30 =            ru|        delta
// v31 =             -|           io
//
//  w6, w7 = plot x0 x1 
// w10             = global plot x
// w15,w16 = plot x2 x3
// x23 = backup address for end of line
// w24 = pixel plot reg for end of line subroutines
// x25 = pixel line position address
// x26 = pixel array start address

.text

.global CalcMandelbrotASM_FracNEON_dp_v82_mt2
CalcMandelbrotASM_FracNEON_dp_v82_mt2:

// to allow use of all regs and call further subroutines,
// backup x17...x30, d8...d15 due to calling conventions
// (x17 for align reasons and x18 due to system variations)
 STP x17,x18,[sp,#-224]!
 STP x19,x20,[sp,#  16]
 STP x21,x22,[sp,#  32]
 STP x23,x24,[sp,#  48]
 STP x25,x26,[sp,#  64]
 STP x27,x28,[sp,#  80]
 STP x29,x30,[sp,#  96]
 STP  d8, d9,[sp,# 112]
 STP d10,d11,[sp,# 128]
 STP d12,d13,[sp,# 144]
 STP d14,d15,[sp,# 160]
 STP x1,x2,  [sp,# 176] // X1 global address for iteration counter
                        // X2 global address for y 

 MOV X26,X0             // use X26 for pixel_buffer start address as X0 is used elsewhere
 MOV X13,X2             // backup address of global iteration counter as X2 is used elsewhere  
 MOV X5, X3             // set first y in X5 as X3 is used elsewhere
 MOV X4,#0              // init local iteration counter
 LDR X14,=frac_data     // pointer to fractal data

   FMOV v28.2d,#4.0            // diverge limit = 4.0 for each pixel
   MOVI v29.2s,#2              // counter sub => 2 iterations per loop  
   MOV W12,#(max_iterations-1) // init maximum iterations - 1

   LDP D0,D1,[X14]       // ro (D0), ru (D20)
   LDR D2,[X14,#24]      // fractal size
   INS v30.d[1],v1.d[0]  // save ru in v30.d[1] for later ***
   FSUB D0,D0,D1
   FDIV D0,D0,D2         // delta_sp = (ro-ru)/fractal_size
   INS v30.d[0],v0.d[0]  // save delta in v30.d[0] for later ***

   y_loop:
      MOV W6,#fractal_size  // get pixel_buffer_width
      MUL W6,W6,W5          // pixel_buffer_width * y_counter 
      ADD X25,X26,W6,UXTW#2 // Line position = pixel_buffer + (pixel_buffer_width * y)<<2

      LDR D1,[X14,#16]      // io
      UCVTF D2,W5           // get y 
      INS v0.d[0],v30.d[0]  // get delta 
      FMUL D2,D2,D0         // y*delta         
      FSUB D1,D1,D2         // b0 = io - y * delta
      INS v31.d[0],v1.d[0]  // save io in v31.d[0] for later ***
      DUP v27.2d,v1.d[0]    // save io_save_sp = b0=b1=b2... in v27

      MOV  W6, #0           // plot pixel 0
      MOV  W7, #1           // plot pixel 1
      MOV  W15,#2           // plot pixel 2
      MOV  W16,#3           // plot pixel 3
      MOV  W10,#3           // plot pixel global init 
  
      LDP  q1,q2,[X14,#32]      // v1 =  0.0      1.0
                                // v2 =  2.0      3.0
      DUP  v3.2d,v30.d[1]       // v3 =   ru       ru    
      FMUL v0.2d,v1.2d,v30.d[0] // v0 =  d*x      d*(x+1)
      FADD v0.2d,v0.2d,v3.2d    // v0 = ru+d*x    ru+d*(x+1)
      FMUL v9.2d,v2.2d,v30.d[0] // v9 =  d*(x+2)  d*(x+3)
      FADD v9.2d,v9.2d,v3.2d    // v9 = ru+d*x+2) ru+d*(x+2)

      MOV  v8.16B,v0.16B        // save a0,a1
      MOV v17.16B,v9.16B        // save a2,a3

      MOV  v1.16B,v27.16B       // init y01 = b0
      MOV v10.16B,v27.16B       // init y23 = b0
      DUP  v7.2s,W12            // init iteration counters p01 = max_iter-1
      DUP v16.2s,W12            // init iteration counters p23 = max_iter-1

      MOV W11,#0                // init end of line flags for each pixel

      iteration_loop:
      iteration_loop_continue_01:
         //1.Iteration pixel 01
         FMUL  v2.2d, v0.2d, v0.2d   // x*x
         FMUL  v3.2d, v1.2d, v1.2d   // y*y
         FMUL  v4.2d, v0.2d, v1.2d   // x*y
         FADD  v5.2d, v2.2d, v3.2d   // x*x + y*y *** backup in v5
         FSUB  v0.2d, v2.2d, v3.2d   // x_new = x*x - y*y
         FADD  v1.2d, v4.2d, v4.2d   // y_new = 2*x*y
         FADD  v0.2d, v0.2d, v8.2d   // x_new = x_new + a01
         FADD  v1.2d, v1.2d,v27.2d   // y_new = y_new + b
         //1.Iteration pixel 23
         FMUL v11.2d, v9.2d, v9.2d   // x*x
         FMUL v12.2d,v10.2d,v10.2d   // y*y
         FMUL v13.2d, v9.2d,v10.2d   // x*y
         FADD v14.2d,v11.2d,v12.2d   // x*x + y*y *** backup in v14
         FSUB  v9.2d,v11.2d,v12.2d   // x_new = x*x - y*y
         FADD v10.2d,v13.2d,v13.2d   // y_new = 2*x*y
         FADD  v9.2d, v9.2d,v17.2d   // x_new = x_new + a23
         FADD v10.2d,v10.2d,v27.2d   // y_new = y_new + b
         //2.Iteration pixel 01
         FMUL  v2.2d, v0.2d, v0.2d   // x*x
         FMUL  v3.2d, v1.2d, v1.2d   // y*y
         FMUL  v4.2d, v0.2d, v1.2d   // x*y
         FADD  v6.2d, v2.2d, v3.2d   // x*x + y*y *** backup in v6
         FSUB  v0.2d, v2.2d, v3.2d   // x_new = x*x - y*y
         FADD  v1.2d, v4.2d, v4.2d   // y_new = 2*x*y
         FCMGT v2.2d, v6.2d,v28.2d   // > 4 ?
         FADD  v0.2d, v0.2d, v8.2d   // x_new = x_new + a01
         SUB   v7.2s, v7.2s,v29.2s   // decrement all pixel counter p01 by 2 
         FADD  v1.2d, v1.2d,v27.2d   // y_new = y_new + b
         UMAXV s3,v2.4s
         UMAXV s4,v7.4s
         MOV   W0,v3.s[0]
         MOV   W1,v4.s[0]
         TST W0,W0                 // faster than TBNZ here...check again later
         B.MI diverged_01          // 1 or more pixel diverged
         TST W1,W1                 // faster than TBNZ here...check again later
         B.MI max_iterations_01    // 1 or more pixel reached max iter
      iteration_loop_continue_23:
         //2.Iteration pixel 23
         FMUL v11.2d, v9.2d, v9.2d   // x*x
         FMUL v12.2d,v10.2d,v10.2d   // y*y
         FMUL v13.2d, v9.2d,v10.2d   // x*y
         FADD v15.2d,v11.2d,v12.2d   // x*x + y*y *** backup in v15
         FSUB  v9.2d,v11.2d,v12.2d   // x_new = x*x - y*y
         FADD v10.2d,v13.2d,v13.2d   // y_new = 2*x*y
         FCMGT v11.2d,v15.2d,v28.2d  // > 4 ?
         FADD  v9.2d, v9.2d,v17.2d   // x_new = x_new + a23
         SUB  v16.2s,v16.2s,v29.2s   // decrement all pixel counter p23 by 2 
         FADD v10.2d,v10.2d,v27.2d   // y_new = y_new + b
         UMAXV s12,v11.4s
         UMAXV s13,v16.4s
         MOV   W28,v12.s[0]
         MOV   W29,v13.s[0]
         TST W28,W28             // faster than TBNZ here...check again later
         B.MI diverged_23        // 1 or more pixel diverged
         TST W29,W29             // faster than TBNZ here...check again later
         B.MI max_iterations_23  // 1 or more pixel reached max iter
      B iteration_loop
 
      //############ Pixel 0 Diverge Handling ############
      diverged_01:
      FCMP D6,D28                // > 4 ? pixel 0 diverged  v6.d[0]
      B.LE diverged_1            // not ? -> check pixel 1
	  
      UMOV  W0,v7.s[0]           // get iterations pixel 0  v7.s[0]
      FCMP  D5,D28               // > 4 ? Diverged at 1st ? v5.d[0]
      CSINC W0,W0,W0,LE          // if => inc counter

        SUB   W2,W12,W0          // correct counter
        ADD   W10,W10,#1         // inc x
        ADD   W4,W4, W2          // add to global iteration counter

        LSL   W2,W2,#colour_shift// some colouring
        STR   W2,[X25,W6,UXTW#2] // write pixel to buffer

        CMP   W10,#fractal_size  // end of line reached ?
        B.LT skip_endofline_p0
          ORR W11,W11,#0b0001    // set end of line flags for pixel 0
          B   diverged_1       // but still check if also another diverged
        skip_endofline_p0:

        INS   v7.s[0],W12        // transfer counter init to v7.s[0]
        MOV   W6,W10             // save new x in new plot x1
        UCVTF D4,W10
        INS   v2.d[0],v30.d[0]   // D2 = delta v2 not needed anymore 
        INS   v3.d[0],v30.d[1]   // D3 =    ru v3 not needed anymore 
        FMUL  D2,D2,D4           // new x0 = x0*d
        FADD  D2,D2,D3           // new x0 = x0*d + ru
        INS   v0.d[0], v2.d[0]   // place new x0 in correct vector v0.d[0]
        INS   v8.d[0], v2.d[0]   // new x0 also = new a0
        INS   v1.d[0],v31.d[0]   // io => get saved y1=y0
        //just continue checking next pixel

      //############ Pixel 1 Diverge Handling ############
      diverged_1:
      INS  v2.d[0],v6.d[1]       // get pixel 1 from v6.d[1]
      FCMP D2,D28                // > 4 ? pixel 1 diverged  v6.d[1]
      B.LE end_div_check_01

      INS   v2.d[0],v5.d[1]      // get diverge value from 1st iter
      UMOV  W0,v7.s[1]           // get iterations pixel 1  v7.s[1]
      FCMP  D2,D28               // > 4 ? Diverged at 2nd ? v5.d[1]
      CSINC W0,W0,W0,LE          // if => inc counter

        SUB   W2,W12,W0          // correct counter
        ADD   W10,W10,#1         // inc x
        ADD   W4,W4, W2          // add to global iteration counter

        LSL   W2,W2,#colour_shift// some colouring
        STR   W2,[X25,W7,UXTW#2] // write pixel to buffer

        CMP   W10,#fractal_size  // end of line reached ?
        B.LT skip_endofline_p1
          ORR W11,W11,#0b0010    // set end of line flags for pixel 1
          B   end_div_check_01
        skip_endofline_p1:

        INS   v7.s[1],W12        // transfer counter init to v7.s[1]
        MOV   W7,W10             // save new x in new plot x1
        UCVTF D4,W10
        INS   v2.d[0],v30.d[0]   // D2 = delta v2 not needed anymore 
        INS   v3.d[0],v30.d[1]   // D3 =    ru v3 not needed anymore 
        FMUL  D2,D2,D4           // new x1 = x1*d
        FADD  D2,D2,D3           // new x1 = ru + x1*d
        INS   v0.d[1], v2.d[0]   // place new x1 in correct vector v0.d[1]
        INS   v8.d[1], v2.d[0]   // new x1 also = new a1    
        INS   v1.d[1],v31.d[0]   // io => get saved y1=y0
        //just continue checking next pixel

      end_div_check_01:

    UMAXV S2,v7.4s                     // check if max iter p01
    MOV   W0,v2.s[0]
    TBNZ  W0,31,max_iterations_01      // 1 or more pixel reached also max iter
    CBZ W11,iteration_loop_continue_23 // if not, was end of line reached ?
    B end_of_line

      //############ Pixel 0 Max Iterations Handling ############
      max_iterations_01: 
      MOV W0,v7.s[0]               // check if pixel 0 reach max iter
      TBZ  W0,31,max_iterations_1  // not ? -> check pixel 1

      ADD W10,W10,#1           // inc x
      ADD W4,W4,#max_iterations// add to global iteration counter
      CMP W10,#fractal_size    // end of line reached ?
      B.LT max_skip_endofline_p0
        ORR W11,W11,#0b0001    // set end of line flags for pixel 0
        B   max_iterations_1   // but still check if also another reached max iter
      max_skip_endofline_p0:

      INS   v7.s[0],W12        // transfer counter init to v7.s[0]
      MOV   W6,W10             // save new x in new plot x1
      UCVTF D4,W10
      INS   v2.d[0],v30.d[0]   // D2 = delta v2 not needed anymore 
      INS   v3.d[0],v30.d[1]   // D3 =    ru v3 not needed anymore 
      FMUL  D2,D2,D4           // new x0 = x0*d
      FADD  D2,D2,D3           // new x0 = x0*d + ru
      INS   v0.d[0], v2.d[0]   // place new x0 in correct vector v0.d[0]
      INS   v8.d[0], v2.d[0]   // new x0 also = new a0
      INS   v1.d[0],v31.d[0]   // io => get saved y1=y0
      //just continue checking next pixel

      //############ Pixel 1 Max Iterations Handling ############
      max_iterations_1:
      MOV W0,v7.s[1]              // check if pixel 1 reach max iter
      TBZ W0,31,end_max_iter_01

      ADD W10,W10,#1           // inc x
      ADD W4,W4,#max_iterations// add to global iteration counter
      CMP W10,#fractal_size    // end of line reached ?
      B.LT max_skip_endofline_p1
        ORR W11,W11,#0b0010    // set end of line flags for pixel 1
        B   end_max_iter_01
      max_skip_endofline_p1:

      INS   v7.s[1],W12        // transfer counter init to v7.s[1]
      MOV   W7,W10             // save new x in new plot x1
      UCVTF D4,W10
      INS   v2.d[0],v30.d[0]   // D2 = delta v2 not needed anymore 
      INS   v3.d[0],v30.d[1]   // D3 =    ru v3 not needed anymore 
      FMUL  D2,D2,D4           // new x1 = x1*d
      FADD  D2,D2,D3           // new x1 = ru + x1*d
      INS   v0.d[1], v2.d[0]   // place new x1 in correct vector v0.d[1]
      INS   v8.d[1], v2.d[0]   // new x1 also = new a1    
      INS   v1.d[1],v31.d[0]   // io => get saved y1=y0
      //just continue checking next pixel

   end_max_iter_01:
   CBZ W11,iteration_loop_continue_23
   B end_of_line

      //############ Pixel 2 Diverge Handling ############
      diverged_23:
      FCMP D15,D28               // > 4 ? pixel 2 diverged  v15.d[0]
      B.LE diverged_3
	  
      UMOV  W0,v16.s[0]          // get iterations pixel 2  v16.s[0]
      FCMP  D14,D28              // > 4 ? Diverged at 1st ? v14.d[0]
      CSINC W0,W0,W0,LE          // if => inc counter

        SUB   W2,W12,W0          // correct counter
        ADD   W10,W10,#1         // inc x
        ADD   W4,W4, W2          // add to global iteration counter

        LSL   W2,W2,#colour_shift// some colouring
        STR   W2,[X25,W15,UXTW#2]// write pixel to buffer

        CMP   W10,#fractal_size  // end of line reached ?
        B.LT skip_endofline_p2
          ORR W11,W11,#0b0100    // set end of line flags for pixel 2
          B   diverged_3         // but still check if also another diverged
        skip_endofline_p2:

        INS   v16.s[0],W12       // transfer counter init to v16.s[0]
        MOV   W15,W10            // save new x in new plot x2
        UCVTF D13,W10
        INS   v11.d[0],v30.d[0]  // D11 = delta v11 not needed anymore 
        INS   v12.d[0],v30.d[1]  // D12 =    ru v12 not needed anymore 
        FMUL  D11,D11,D13        // new x2 = x2*d
        FADD  D11,D11,D12        // new x2 = x2*d + ru
        INS    v9.d[0],v11.d[0]  // place new x2 in correct vector v9.d[0]
        INS   v17.d[0],v11.d[0]  // new x2 also = new a2
        INS   v10.d[0],v31.d[0]  // io => get saved y2=y3

      //############ Pixel 3 Diverge Handling ############
      diverged_3:
      INS  v11.d[0],v15.d[1]     // get pixel 3 from v15.d[1]
      FCMP D11,D28               // > 4 ? pixel 3 diverged  v15.d[1]
      B.LE end_div_check_23 

      INS   v11.d[0],v14.d[1]    // get diverge value from 1st iter
      UMOV  W0,v16.s[1]          // get iterations pixel 3  v16.s[1]
      FCMP  D11,D28              // > 4 ? Diverged at 1st ? v14.d[1]
      CSINC W0,W0,W0,LE          // if => inc counter

        SUB   W2,W12,W0          // correct counter
        ADD   W10,W10,#1         // inc x
        ADD   W4,W4, W2          // add to global iteration counter

        LSL   W2,W2,#colour_shift// some colouring
        STR   W2,[X25,W16,UXTW#2]// write pixel to buffer

        CMP   W10,#fractal_size  // end of line reached ?
        B.LT skip_endofline_p3
          ORR W11,W11,#0b1000    // set end of line flags for pixel 1
          B   end_div_check_23   // but still check if also another diverged
        skip_endofline_p3:

        INS   v16.s[1],W12       // transfer counter init to v16.s[1]
        MOV   W16,W10            // save new x in new plot x3
        UCVTF D13,W10
        INS   v11.d[0],v30.d[0]  // D11 = delta v11 not needed anymore 
        INS   v12.d[0],v30.d[1]  // D12 =    ru v12 not needed anymore 
        FMUL  D11,D11,D13        // new x3 = x3*d
        FADD  D11,D11,D12        // new x3 = x3*d + ru
        INS    v9.d[1],v11.d[0]  // place new x3 in correct vector v9.d[1]
        INS   v17.d[1],v11.d[0]  // new x3 also = new a3
        INS   v10.d[1],v31.d[0]  // io => get saved y2=y3
        //just continue checking next pixel

      end_div_check_23:

    UMAXV S11,v16.4s                   // check if max iter p23
    MOV   W0,v11.s[0]
    TBNZ  W0,31,max_iterations_23      // 1 or more pixel reached also max iter
    CBZ W11,iteration_loop_continue_01 // if not, was end of line reached ?
    B end_of_line

      //############ Pixel 2 Max Iterations Handling ############
      max_iterations_23: 
      UMOV W0,v16.s[0]              // check if pixel 2 reach max iter
      TBZ  W0,31,max_iterations_3

      ADD W10,W10,#1           // inc x
      ADD W4,W4,#max_iterations// add to global iteration counter
      CMP W10,#fractal_size    // end of line reached ?
      B.LT max_skip_endofline_p2
        ORR W11,W11,#0b0100    // set end of line flags for pixel 2
        B   max_iterations_3   // but still check if also another reached max iter
      max_skip_endofline_p2:

      INS   v16.s[0],W12       // transfer counter init to v16.s[0]
      MOV   W15,W10            // save new x in new plot x2
      UCVTF D13,W10
      INS   v11.d[0],v30.d[0]  // D11 = delta v11 not needed anymore 
      INS   v12.d[0],v30.d[1]  // D12 =    ru v12 not needed anymore 
      FMUL  D11,D11,D13        // new x2 = x2*d
      FADD  D11,D11,D12        // new x2 = x2*d + ru
      INS    v9.d[0],v11.d[0]  // place new x2 in correct vector v9.d[0]
      INS   v17.d[0],v11.d[0]  // new x2 also = new a2
      INS   v10.d[0],v31.d[0]  // io => get saved y2=y3

      //############ Pixel 3 Max Iterations Handling ############
      max_iterations_3:
      UMOV W0,v16.s[1]             // check if pixel 3 reach max iter
      TBZ  W0,31,end_max_iter_23 

      ADD W10,W10,#1           // inc x
      ADD W4,W4,#max_iterations// add to global iteration counter
      CMP W10,#fractal_size    // end of line reached ?
      B.LT max_skip_endofline_p3
        ORR W11,W11,#0b1000    // set end of line flags for pixel 3
        B   end_max_iter_23 
      max_skip_endofline_p3:

      INS   v16.s[1],W12       // transfer counter init to v16.s[1]
      MOV   W16,W10            // save new x in new plot x3
      UCVTF D13,W10
      INS   v11.d[0],v30.d[0]  // D11 = delta v11 not needed anymore 
      INS   v12.d[0],v30.d[1]  // D12 =    ru v12 not needed anymore 
      FMUL  D11,D11,D13        // new x3 = x3*d
      FADD  D11,D11,D12        // new x3 = x3*d + ru
      INS    v9.d[1],v11.d[0]  // place new x3 in correct vector v9.d[1]
      INS   v17.d[1],v11.d[0]  // new x3 also = new a3
      INS   v10.d[1],v31.d[0]  // io => get saved y2=y3
      //just continue checking next pixel

      end_max_iter_23:
      CBZ W11,iteration_loop_continue_01

      //############ End of Line Handling  ############
//  v0 = x01
//  v1 = y01
//  v2 = calc p01 (x*x),...
//  v3 = calc p01 (y*y),...
//  v4 = calc p01 (x*x),...
//  v5 = x*x+y*y save iter p01 0
//  v6 = x*x+y*y save iter p01 1
//  v7 = counter p01
//  v8 = a01

//  v9 = x23
// v10 = y23
// v11 = calc p23 (x*x),...
// v12 = calc p23 (y*y),...
// v13 = calc p23 (x*x),...
// v14 = x*x+y*y save iter p23 0
// v15 = x*x+y*y save iter p23 1
// v16 = counter p23
// v17 = a23

      end_of_line:
  
      //############ End of Line Pixel 0 Handling  ############
      check_end_of_line_pixel_0:
      TBNZ W11,#0,check_end_of_line_pixel_1
        STP Q0,Q1,[sp,#192]  //backup reg v0,v1 (x0123,y0123)
        INS v31.d[0],v8.d[0] //get a0     
        MOV       W0,v7.s[0] //get iterations p0
        MOV W24,W6           //get plot_x
        BL end_of_line_01_iter
        LDP Q0,Q1,[sp,#192]  //restore reg v0,v1 (x0123,y0123)

      //############ End of Line Pixel 1 Handling  ############
      check_end_of_line_pixel_1:
      TBNZ W11,#1,check_end_of_line_pixel_2
        STP Q0,Q1,[sp,#192]  //backup reg v0,v1 (x0123,y0123)
        INS  v0.d[0],v0.d[1] //get x1
        INS  v1.d[0],v1.d[1] //get y1
        INS v31.d[0],v8.d[1] //get a1
        MOV       W0,v7.s[1] //get iterations p1
        MOV W24,W7           //get plot_x
        BL end_of_line_01_iter
        LDP Q0,Q1,[sp,#192]  //restore reg v0,v1 (x0123,y0123)

      //############ End of Line pixel 2 Handling  ############
      check_end_of_line_pixel_2:
      TBNZ W11,#2,check_end_of_line_pixel_3
        INS  v0.d[0], v9.d[0] //get x2
        INS  v1.d[0],v10.d[0] //get y2
        INS v31.d[0],v17.d[0] //get a2
        INS  v5.d[0],v14.d[0] //get diverge backup
        MOV       W0,v16.s[0] //get iterations p2
        MOV W24,W15           //get plot_x
        BL end_of_line_23_iter

      //############ End of Line pixel 3 Handling  ############
      check_end_of_line_pixel_3:
      TBNZ W11,#3,next_line
        INS  v0.d[0], v9.d[1] //get x3
        INS  v1.d[0],v10.d[1] //get y3
        INS v31.d[0],v17.d[1] //get a3
        INS  v5.d[0],v14.d[1] //get diverge backup
        MOV       W0,v16.s[1] //get iterations p3
        MOV W24,W16           //get plot_x
        BL end_of_line_23_iter

   next_line:

   // atomic update of global y_line counter
.if armv8_2 == true
   MOV     W6,#1
   LDADDAL W6,W5,[X13]   //[X13], [X13] = [X13] + 1
   ADD     W5,W5,#1      //W5=W5+1 for CMP check below
.else
   try_again_update_global_y_line_counter:
     LDAXR W6,[X13]        //try LDXR or LDAXR
     ADD   W5,W6,#1
     STLXR W7,W5,[X13]     //try STXR or STLXR
   CBNZ W7,try_again_update_global_y_line_counter
.endif
   
   CMP W5,#fractal_size    //may be check can already be done within the atomic update, check performance
   B.LO y_loop             //use lower check due to overflow of other cores
   
// prepare to return to C routine
// restore x17...x30, d8...d15
   LDP x17,x18,[sp],#16
   LDP x19,x20,[sp],#16
   LDP x21,x22,[sp],#16
   LDP x23,x24,[sp],#16
   LDP x25,x26,[sp],#16
   LDP x27,x28,[sp],#16
   LDP x29,x30,[sp],#16
   LDP  d8, d9,[sp],#16
   LDP d10,d11,[sp],#16
   LDP d12,d13,[sp],#16
   LDP d14,d15,[sp],#16
   LDP x1, x2, [sp],#48  //get address for iter count storage

// atomic update of global iteration counter
.if armv8_2 == true
   LDADDAL X4,XZR,[X1]     //XZR(not needed) = [X1], [X1] = [X1] + X4
.else
   try_again_update_global_iteration_counter:
     LDAXR X9,[X1]         //try LDXR or LDAXR
     ADD   X6,X9,X4
     STLXR W7,X6,[X1]      //try STXR or STLXR
   CBNZ W7,try_again_update_global_iteration_counter
.endif
      
RET

 //subroutines
   end_of_line_01_iter:
      iterate_pixel_01:
         FMUL  D2, D0, D0 // x*x
         FMUL  D3, D1, D1 // y*y
         FMUL  D4, D0, D1 // x*y
         SUB   W0, W0, #1 // decrement iteration counter
         FADD  D6, D2, D3 // x*x + y*y *** backup in v6
         FSUB  D0, D2, D3 // x_new = x*x - y*y
         FADD  D1, D4, D4 // y_new = 2*x*y
         FCMP  D6,D28     // > 4 ?
         FADD  D0, D0,D31 // x_new = x_new + a01
         FADD  D1, D1,D27 // y_new = y_new + b
         B.GT  end_pixel_01_diverged
       TBZ W0,31,iterate_pixel_01
       ADD W4,W4,#max_iterations // add to global iteration counter
     RET
       end_pixel_01_diverged:
       SUB W2,W12,W0             // correct counter
       ADD W4,W4, W2             // add to global iteration counter
       LSL W2,W2,#colour_shift   // some colouring
       STR W2,[X25,W24,UXTW#2]   // write pixel to buffer
     RET

   end_of_line_23_iter:
      TST W11,#0b1100            // if exit from p01 then one iteration already done !
      B.EQ at_first_one_iter_already_done_23
      iterate_pixel_23:
         FMUL  D2, D0, D0 // x*x
         FMUL  D3, D1, D1 // y*y
         FMUL  D4, D0, D1 // x*y
         FADD  D5, D2, D3 // x*x + y*y
         FSUB  D0, D2, D3 // x_new = x*x - y*y
         FADD  D1, D4, D4 // y_new = 2*x*y
         FADD  D0, D0,D31 // x_new = x_new + a23
         FADD  D1, D1,D27 // y_new = y_new + b
         at_first_one_iter_already_done_23:
         SUB   W0, W0, #2 // decrement iteration counter
         FMUL  D2, D0, D0 // x*x
         FMUL  D3, D1, D1 // y*y
         FMUL  D4, D0, D1 // x*y
         FADD  D6, D2, D3 // x*x + y*y
         FSUB  D0, D2, D3 // x_new = x*x - y*y
         FADD  D1, D4, D4 // y_new = 2*x*y
         FCMP  D6,D28     // > 4 ?
         FADD  D0, D0,D31 // x_new = x_new + a23
         FADD  D1, D1,D27 // y_new = y_new + b
         B.GT  end_pixel_23_diverged
       TBZ W0,31,iterate_pixel_23
       ADD W4,W4,#max_iterations // add to global iteration counter
     RET
       end_pixel_23_diverged:
       FCMP  D5,D28      // > 4 ? Diverged at 1st ?
       CSINC W0,W0,W0,LE // if => inc counter
       SUB   W2,W12,W0   // correct counter
       ADD   W4,W4, W2   // add to global iteration counter
       LSL W2,W2,#colour_shift   // some colouring
       STR W2,[X25,W24,UXTW#2]   // write pixel to buffer
     RET

.data
.align 4
frac_data:
.double -0.1450 //+ 0 ro
.double -0.1750 //+ 8 ru
.double -1.0200 //+16 io
.double   600.0 //+24 fractal_size
.double    0.0  //+32
.double    1.0
.double    2.0
.double    3.0
