//---------------------------------------------------------------
// FracNEON_sp_v8a_mt1 Assembler routine routine by Michael Kübel
//---------------------------------------------------------------
// extern "C" uint64_t CalcMandelbrotASM_FracNEON_sp_v8a_mt1(const uint32_t* p,const uint64_t* iterations,const uint32_t* y,const uint32_t y_start);

fractal_size   =  600 // size of fractal in pixels
max_iterations = 4096 // maximum iterations
colour_shift   =    3 // shift colours left

.text

.global CalcMandelbrotASM_FracNEON_sp_v8a_mt1
CalcMandelbrotASM_FracNEON_sp_v8a_mt1:

// to allow use of all regs and call further subroutines,
// backup x17...x30, d8...d15 due to calling conventions
// (x17 for align reasons and x18 due to system variations)
 STP x17,x18,[sp,#-224]!
 STP x19,x20,[sp,#  16]
 STP x21,x22,[sp,#  32]
 STP x23,x24,[sp,#  48]
 STP x25,x26,[sp,#  64]
 STP x27,x28,[sp,#  80]
 STP x29,x30,[sp,#  96]
 STP  d8, d9,[sp,# 112]
 STP d10,d11,[sp,# 128]
 STP d12,d13,[sp,# 144]
 STP d14,d15,[sp,# 160]
 STP x1,x2,  [sp,# 176] // X1 global address for iteration counter
                        // X2 global address for y 

 MOV X26,X0             // use X26 for pixel_buffer start address as X0 is used elsewhere
 MOV X13,X2             // backup address of global iteration counter as X2 is used elsewhere  
 MOV X27,X3             // set first y in X5 as X3 is used elsewhere
 MOV X4,#0              // init local iteration counter
 LDR X28,=frac_data     // pointer to fractal data
      
   //*** S20 = 'ru' will be global constant *** 
   LDP S0,S20,[X28]      // ro (S0), ru (S20)
   LDR S2,[X28,#12]      // fractal size
   FSUB S0, S0,S20
   //*** S21 = 'delta_sp' will be global constant ***
   FDIV S21,S0, S2       // delta_sp = (ro-ru)/fractal_size
   FMOV v15.4s,#4.0      // limit = 4.0 for each pixel

   MOVI v14.4s,#4               // init sub value = 4 for each pixel (4 times loop unrolling)   
   MOV W12,#(max_iterations-1)  // init maximum iterations - 1

   y_loop:

      MOV W6,#fractal_size  // get pixel_buffer_width
      MUL W6,W6,W27         // pixel_buffer_width * y_counter 
      ADD X25,X26,W6,UXTW#2 // Line position = pixel_buffer + (pixel_buffer_width * y)<<2
   
      LDR S0,[X28,#8]      // io

      UCVTF S2,W27         // get y  
      FMUL  S2, S2,S21     // use FMSUB later	  
      //*** S22 = 'io_save_sp' will be global constant *** 
      FSUB  S22,S0, S2     // b0 = io - current_y * delta
      DUP v5.4s,v22.s[0]   // save io_save_sp = b0=b1=b2=b3 in Q5

      MOV W6 ,#0           // starting line counter pixel 1
      MOV W7, #1           // starting line counter pixel 2
      MOV W8, #2           // starting line counter pixel 3
      MOV W9, #3           // starting line counter pixel 4
      MOV W10,W9           // init line counter global
  
      LDR  q1,[X28,#16]         // v1 =  0.0      1.0       2.0       3.0
      DUP  v0.4s,v20.s[0]       // v0 =   ru       ru        ru        ru
      FMUL v1.4s,v1.4s,v21.s[0] // v1 =  d*x      d*(x+1)   d*(x+2)   d*(x+3)
      FADD v0.4s,v0.4s,v1.4s    // v0 = ru+d*x  ru+d*(x+1) ru+d*(x+2)  ru+d*(x+3)
                                // x  = a ...*** use FMLA later...
	  
      MOV v4.16B,v0.16B      // save a0,a1,a2,a3
      MOV v1.16B,v5.16B      // y = b0

      DUP  v2.4s,W12         // init single iteration counters = max_iter-1 ***use vector

      MOV X11,#0             // init end of line flags for each pixel

      iteration_loop:
         //1.Iteration
         FMUL  v3.4s, v0.4s, v0.4s   // x*x
         FMUL  v8.4s, v1.4s, v1.4s   // y*y
         FMUL v12.4s, v0.4s, v1.4s   // x*y
         FADD v10.4s, v3.4s, v8.4s   // x*x + y*y
         FSUB  v9.4s, v3.4s, v8.4s   // x_new = x*x - y*y
         FADD v12.4s,v12.4s,v12.4s   // 2*x*y
         FADD  v0.4s, v9.4s, v4.4s   // x_new = x_new + a
         FADD  v1.4s,v12.4s, v5.4s   // y_new = y_new + b

         //2.Iteration
         FMUL  v3.4s, v0.4s, v0.4s   // x*x
         FMUL  v8.4s, v1.4s, v1.4s   // y*y
         FMUL v12.4s, v0.4s, v1.4s   // x*y
         FADD v11.4s, v3.4s, v8.4s   // x*x + y*y
         FSUB  v9.4s, v3.4s, v8.4s   // x_new = x*x - y*y
         FADD v12.4s,v12.4s,v12.4s   // 2*x*y
         FADD  v0.4s, v9.4s, v4.4s   // x_new = x_new + a
         FADD  v1.4s,v12.4s, v5.4s   // y_new = y_new + b

         //3.Iteration
         FMUL  v3.4s, v0.4s, v0.4s   // x*x
         FMUL  v8.4s, v1.4s, v1.4s   // y*y
         FMUL v12.4s, v0.4s, v1.4s   // x*y
         FADD  v6.4s, v3.4s, v8.4s   // x*x + y*y
         FSUB  v9.4s, v3.4s, v8.4s   // x_new = x*x - y*y
         FADD v12.4s,v12.4s,v12.4s   // 2*x*y
         FADD  v0.4s, v9.4s, v4.4s   // x_new = x_new + a
         FADD  v1.4s,v12.4s, v5.4s   // y_new = y_new + b

         //4.Iteration
         FMUL  v3.4s, v0.4s, v0.4s   // x*x
         FMUL  v8.4s, v1.4s, v1.4s   // y*y
         FMUL v12.4s, v0.4s, v1.4s   // x*y
         FADD  v7.4s, v3.4s, v8.4s   // x*x + y*y
         FSUB  v9.4s, v3.4s, v8.4s   // x_new = x*x - y*y
         FADD v12.4s,v12.4s,v12.4s   // 2*x*y
         FCMGT v3.4s, v7.4s,v15.4s   // > 4 ?
         FADD  v0.4s, v9.4s, v4.4s   // x_new = x_new + a
         FADD  v1.4s,v12.4s, v5.4s   // y_new = y_new + b

         SUB   v2.4s, v2.4s,v14.4s   // decrement pixel counters
 
         UMAXV s17,v3.4s             // that order here seems to be the fastest
         UMAXV s16,v2.4s 
         MOV   w17,v17.s[0]
         MOV   w16,v16.s[0]

         TST W17,W17            // faster than TBNZ here
         B.MI  diverged_1234    // 1 or more pixel diverged
         TST W16,W16            // faster than TBNZ here
      B.PL iteration_loop
      B.MI max_iterations_1234  // 1 or more pixel reached max iter
 
      //############ Pixel 1 Diverge Handling ############

      diverged_1234:
      FCMP S7,S15                // > 4 ? pixel 1 diverged   v7.4s[0]
      B.LE diverged_234          // not ? -> check pixel 2,3,4
  
      UMOV  W0,v2.s[0]         // get iterations pixel 1   v2.4s[0]	 
      FCMP  S6,S15             // > 4 ? Diverged at 3rd ?  v6.4s[0]
      CSINC W0,W0,W0,LE        // if => inc counter
      FCMP  S11,S15            // > 4 ? Diverged at 2nd ? v11.4s[0]
      CSINC W0,W0,W0,LE        // if => inc counter
      FCMP  S10,S15            // > 4 ? Diverged at 1st ? v10.4s[0]
      CSINC W0,W0,W0,LE        // if => inc counter

        SUB   W2,W12,W0          // correct counter
        ADD   W10,W10,#1         // inc x
        ADD   W4,W4, W2          // add to global iteration counter

        LSL   W2,W2,#colour_shift// some colouring
        STR   W2,[X25,W6,UXTW#2] // write pixel to buffer

        CMP   W10,#fractal_size  // end of line reached ?
        B.LT skip_endofline_p1
          ORR W11,W11,#0b0001    // set end of line flags for pixel 1
          B   diverged_234       // but still check if also another diverged
        skip_endofline_p1:

        INS   v2.s[0],W12        // transfer counter init to v2.s[0]
        MOV   W6,W10             // save new x in new plot_x1
        UCVTF S23,W10
        FMUL  S23,S23,S21        // new x1 = x1*d
        FADD  S23,S23,S20        // new x1 = ru + x1*d
        INS   v0.s[0],v23.s[0]   // place new x1 in correct vector v0.s[0]
        INS   v4.s[0],v23.s[0]   // new x1 also = new 'a'    
        INS   v1.s[0],v22.s[0]   // io_save_sp => get saved y4=y3=y2=y1
 
         //just continue checking pixel 2

      //############ Pixel 2 Diverge Handling ############

      diverged_234:
      INS  v23.s[0],v7.s[1]      // get pixel 2 from v7.4[1]
      FCMP S23,S15               // > 4 ? pixel 2 diverged
      B.LE diverged_34           // not ? -> check pixel 3,4

      INS   v23.s[0], v6.s[1]  // place out of order
      INS   v24.s[0],v11.s[1]
      INS   v25.s[0],v10.s[1]
      UMOV  W0,v2.s[1]         // get iterations pixel 2   v2.4s[1]	 
      FCMP  S23,S15            // > 4 ? Diverged at 3rd ?  v6.4s[1]
      CSINC W0,W0,W0,LE        // if => inc counter
      FCMP  S24,S15            // > 4 ? Diverged at 2nd ? v11.4s[1]
      CSINC W0,W0,W0,LE        // if => inc counter
      FCMP  S25,S15            // > 4 ? Diverged at 1st ? v10.4s[1]
      CSINC W0,W0,W0,LE        // if => inc counter

        SUB   W2,W12,W0          // correct counter
        ADD   W10,W10,#1         // inc x
        ADD   W4,W4, W2          // add to global iteration counter

        LSL   W2,W2,#colour_shift// some colouring
        STR   W2,[X25,W7,UXTW#2] // write pixel to buffer

        CMP   W10,#fractal_size  // end of line reached ?
        B.LT skip_endofline_p2
          ORR W11,W11,#0b0010    // set end of line flags for pixel 2
          B   diverged_34        // but still check if also another diverged
        skip_endofline_p2:

        INS   v2.s[1],W12        // transfer counter init to v2.s[1]
        MOV   W7,W10             // save new x in new plot_x2
        UCVTF S23,W10
        FMUL  S23,S23,S21        // new x2 = x2*d
        FADD  S23,S23,S20        // new x2 = ru + x2*d
        INS   v0.s[1],v23.s[0]   // place new x1 in correct vector v0.s[1]
        INS   v4.s[1],v23.s[0]   // new x2 also = new 'a'    
        INS   v1.s[1],v22.s[0]   // io_save_sp => get saved y4=y3=y2=y1

        //just continue checking pixel 3

      //############ Pixel 3 Diverge Handling ############

      diverged_34:
      INS  v23.s[0],v7.s[2]      // get pixel 3 from v7.4[2]
      FCMP S23,S15               // > 4 ? pixel 3 diverged
      B.LE diverged_4            // not ? -> check pixel 4

      INS   v23.s[0], v6.s[2]
      INS   v24.s[0],v11.s[2]
      INS   v25.s[0],v10.s[2]
      UMOV  W0,v2.s[2]         // get iterations pixel 3   v2.4s[2]	 
      FCMP  S23,S15            // > 4 ? Diverged at 3rd ?  v6.4s[2]
      CSINC W0,W0,W0,LE        // if => inc counter
      FCMP  S24,S15            // > 4 ? Diverged at 2nd ? v11.4s[2]
      CSINC W0,W0,W0,LE        // if => inc counter
      FCMP  S25,S15            // > 4 ? Diverged at 1st ? v10.4s[2]
      CSINC W0,W0,W0,LE        // if => inc counter
 
        SUB   W2,W12,W0          // correct counter
        ADD   W10,W10,#1         // inc x
        ADD   W4,W4, W2          // add to global iteration counter

        LSL   W2,W2,#colour_shift// some colouring
        STR   W2,[X25,W8,UXTW#2] // write pixel to buffer

        CMP   W10,#fractal_size  // end of line reached ?
        B.LT skip_endofline_p3
          ORR W11,W11,#0b0100    // set end of line flags for pixel 3
          B   diverged_4         // but still check if also another diverged
        skip_endofline_p3:

        INS   v2.s[2],W12        // transfer counter init to v2.s[2]
        MOV   W8,W10             // save new x in new plot_x3
        UCVTF S23,W10
        FMUL  S23,S23,S21        // new x3 = x3*d
        FADD  S23,S23,S20        // new x3 = ru + x3*d
        INS   v0.s[2],v23.s[0]   // place new x1 in correct vector v0.s[2]
        INS   v4.s[2],v23.s[0]   // new x3 also = new 'a'    
        INS   v1.s[2],v22.s[0]   // io_save_sp => get saved y4=y3=y2=y1

      //############ Pixel 4 Diverge Handling ############

      diverged_4:
      INS  v23.s[0],v7.s[3]      // get pixel 4 from v7.4[3]
      FCMP S23,S15               // > 4 ? pixel 4 diverged
      B.LE end_diverge_check     // not ? -> continue

      INS   v23.s[0], v6.s[3]
      INS   v24.s[0],v11.s[3]
      INS   v25.s[0],v10.s[3]
      UMOV  W0,v2.s[3]         // get iterations pixel 4   v2.4s[3]	  
      FCMP  S23,S15            // > 4 ? Diverged at 3rd ?  v6.4s[3]
      CSINC W0,W0,W0,LE        // if => inc counter
      FCMP  S24,S15            // > 4 ? Diverged at 2nd ? v11.4s[3]
      CSINC W0,W0,W0,LE        // if => inc counter
      FCMP  S25,S15            // > 4 ? Diverged at 1st ? v10.4s[3]
      CSINC W0,W0,W0,LE        // if => inc counter

        SUB   W2,W12,W0          // correct counter
        ADD   W10,W10,#1         // inc x
        ADD   W4,W4, W2          // add to global iteration counter

        LSL   W2,W2,#colour_shift// some colouring
        STR   W2,[X25,W9,UXTW#2] // write pixel to buffer

        CMP   W10,#fractal_size  // end of line reached ?
        B.LT skip_endofline_p4
          ORR W11,W11,#0b1000    // set end of line flags for pixel 3
          B   end_diverge_check  // but still check if also another diverged
        skip_endofline_p4:

        INS   v2.s[3],W12        // transfer counter init to v2.s[3]
        MOV   W9,W10             // save new x in new plot_x4
        UCVTF S23,W10
        FMUL  S23,S23,S21        // new x4 = x4*d
        FADD  S23,S23,S20        // new x4 = ru + x4*d
        INS   v0.s[3],v23.s[0]   // place new x4 in correct vector v0.s[3]
        INS   v4.s[3],v23.s[0]   // new x4 also = new 'a'    
        INS   v1.s[3],v22.s[0]   // io_save_sp => get saved y4=y3=y2=y1

      end_diverge_check:

      UMAXV s16,v2.4s             // check if v2<0
      MOV   W16,v16.s[0]          // *** use FCMP somehow ?
      TST   W16,W16
      B.MI  max_iterations_1234   // 1 or more pixel reached also max iter

      CBZ   W11,iteration_loop    // if not, was end of line reached ?

      //############ Pixel 1 Max Iterations Handling ############

      max_iterations_1234: 
      MOV  W0,v2.s[0]          // check if pixel 1 reach max iter
      TBZ W0,31,max_iterations_234 // not ? -> check pixel 2,3,4

      ADD W10,W10,#1           // inc x
      ADD W4,W4,#max_iterations// add to global iteration counter
      CMP W10,#fractal_size    // end of line reached ?
      B.LT max_skip_endofline_p1
        ORR W11,W11,#0b0001    // set end of line flags for pixel 1
        B   max_iterations_234 // but still check if also another diverged
      max_skip_endofline_p1:

      MOV   W6,W10             // save new x in new plot_x1
      INS   v2.s[0],W12        // transfer counter init to v2.s[0]
      UCVTF S23,W10
      FMUL  S23,S23,S21        // new x1 = x1*d
      FADD  S23,S23,S20        // new x1 = ru + x1*d
      INS   v0.s[0],v23.s[0]   // place new x1 in correct vector v0.s[0]
      INS   v4.s[0],v23.s[0]   // new x1 also = new 'a'    
      INS   v1.s[0],v22.s[0]   // io_save_sp => get saved y4=y3=y2=y1

      //############ Pixel 2 Max Iterations Handling ############

      max_iterations_234:
      MOV  W0,v2.s[1]          // check if pixel 2 reach max iter
      TBZ W0,31,max_iterations_34 // not ? -> check pixel 3,4

      ADD W10,W10,#1           // inc x
      ADD W4,W4,#max_iterations// add to global iteration counter
      CMP W10,#fractal_size    // end of line reached ?
      B.LT max_skip_endofline_p2
        ORR W11,W11,#0b0010    // set end of line flags for pixel 2
        B   max_iterations_34  // but still check if also another diverged
      max_skip_endofline_p2:

      MOV   W7,W10             // save new x in new plot_x2
      INS   v2.s[1],W12        // transfer counter init to v2.s[1]
      UCVTF S23,W10
      FMUL  S23,S23,S21        // new x2 = x2*d
      FADD  S23,S23,S20        // new x2 = ru + x2*d
      INS   v0.s[1],v23.s[0]   // place new x1 in correct vector v0.s[1]
      INS   v4.s[1],v23.s[0]   // new x2 also = new 'a'    
      INS   v1.s[1],v22.s[0]   // io_save_sp => get saved y4=y3=y2=y1

      //############ Pixel 3 Max Iterations Handling ############

      max_iterations_34:
      MOV  W0,v2.s[2]          // check if pixel 3 reach max iter
      TBZ W0,31,max_iterations_4 // not ? -> check pixel 4

      ADD W10,W10,#1           // inc x
      ADD W4,W4,#max_iterations// add to global iteration counter
      CMP W10,#fractal_size    // end of line reached ?
      B.LT max_skip_endofline_p3
        ORR W11,W11,#0b0100    // set end of line flags for pixel 3
        B   max_iterations_4   // but still check if also another diverged
      max_skip_endofline_p3:
	  
      MOV   W8,W10             // save new x in new plot_x3
      INS   v2.s[2],W12        // transfer counter init to v2.s[2]
      UCVTF S23,W10
      FMUL  S23,S23,S21        // new x3 = x3*d
      FADD  S23,S23,S20        // new x3 = ru + x3*d
      INS   v0.s[2],v23.s[0]   // place new x1 in correct vector v0.s[2]
      INS   v4.s[2],v23.s[0]   // new x3 also = new 'a'    
      INS   v1.s[2],v22.s[0]   // io_save_sp => get saved y4=y3=y2=y1

      //############ Pixel 4 Max Iterations Handling ############

      max_iterations_4:
      MOV  W0,v2.s[3]          // check if pixel 4 reach max iter
      TBZ W0,31,end_max_iter_check // not ? -> end max iter check

      ADD W10,W10,#1           // inc x
      ADD W4,W4,#max_iterations// add to global iteration counter
      CMP W10,#fractal_size    // end of line reached ?
      B.LT max_skip_endofline_p4
        ORR W11,W11,#0b1000    // set end of line flags for pixel 4
        B   end_of_line        // now all are checked for everything
      max_skip_endofline_p4:

      MOV   W9,W10             // save new x in new plot_x4
      INS   v2.s[3],W12        // transfer counter init to v2.s[3]
      UCVTF S23,W10
      FMUL  S23,S23,S21        // new x4 = x4*d
      FADD  S23,S23,S20        // new x4 = ru + x4*d
      INS   v0.s[3],v23.s[0]   // place new x4 in correct vector v0.s[3]
      INS   v4.s[3],v23.s[0]   // new x4 also = new 'a'    
      INS   v1.s[3],v22.s[0]   // io_save_sp => get saved y4=y3=y2=y1

      end_max_iter_check:

      CBZ W11,iteration_loop

      end_of_line:

      MOV v24.16B,v0.16B // save state of v0 for other pixels
      MOV v25.16B,v1.16B // save state of v1 for other pixels
 
      //############ End of Line Pixel 1 Handling  ############
      
      TBNZ W11,#0,check_end_of_line_pixel_2 //test for 0b0001

      MOV W1,v2.s[0]  
      iterate_pixel_1:
         FMUL  s3, s0, s0 // x*x
         FMUL  s8, s1, s1 // y*y
         SUB   W1, W1, #1 // decrement pixel counter
         FMUL s12, s0, s1 // x*y
         FADD  s7, s3, s8 // x*x + y*y
         FSUB  s9, s3, s8 // x_new = x*x - y*y
         FADD s12,s12,s12 // 2*x*y
         FADD  s0, s9, s4 // x_new = x_new + a
         FCMP  s7,s15     // pixel 1 > 4 ? check if pixel 1 diverged
         FADD  s1,s12, s5 // y_new = y_new + b
         B.GT  end_pixel_1_diverged
         TBZ W1,31,iterate_pixel_1

	  //maximum iterations reached
         ADD W4,W4,#max_iterations   // add to global iteration counter
         B check_end_of_line_pixel_2

         end_pixel_1_diverged:
         SUB W2,W12,W1          // correct counter
         ADD W4,W4, W2          // add to global iteration counter
         LSL W2,W2,#colour_shift// some colouring
         STR W2,[X25,W6,UXTW#2] // write pixel to buffer

      //############ End of Line Pixel 2 Handling  ############

      check_end_of_line_pixel_2:

      MOV v0.16B,v24.16B // restore state of v0 for other pixels
      MOV v1.16B,v25.16B // restore state of v1 for other pixels
 	  
      TBNZ W11,#1,check_end_of_line_pixel_3 //test for 0b0010

      INS v0.s[0],v0.s[1] 
      INS v1.s[0],v1.s[1] 
      INS v4.s[0],v4.s[1] 
      MOV W1,v2.s[1]  
      iterate_pixel_2:
         FMUL  s3, s0, s0 // x*x
         FMUL  s8, s1, s1 // y*y
         SUB   W1, W1, #1 // decrement pixel counter
         FMUL s12, s0, s1 // x*y
         FADD  s7, s3, s8 // x*x + y*y
         FSUB  s9, s3, s8 // x_new = x*x - y*y
         FADD s12,s12,s12 // 2*x*y
         FADD  s0, s9, s4 // x_new = x_new + a
         FCMP  s7,s15     // pixel 1 > 4 ? check if pixel 1 diverged
         FADD  s1,s12, s5 // y_new = y_new + b
         B.GT  end_pixel_2_diverged
         TBZ W1,31,iterate_pixel_2
 
         // maximum iterations reached
         ADD W4,W4,#max_iterations   // add to global iteration counter
         B check_end_of_line_pixel_3

         end_pixel_2_diverged:
         SUB W2,W12,W1          // correct counter
         ADD W4,W4, W2          // add to global iteration counter
         LSL W2,W2,#colour_shift// some colouring
         STR W2,[X25,W7,UXTW#2] // write pixel to buffer

      //############ End of Line Pixel 3 Handling  ############

      check_end_of_line_pixel_3:

      MOV v0.16B,v24.16B // restore state of v0 for other pixels
      MOV v1.16B,v25.16B // restore state of v1 for other pixels
 
      TBNZ W11,#2,check_end_of_line_pixel_4 //test for 0b0100
 
      INS v0.s[0],v0.s[2] 
      INS v1.s[0],v1.s[2] 
      INS v4.s[0],v4.s[2] 
      MOV W1,v2.s[2]  
      iterate_pixel_3:
         FMUL  s3, s0, s0 // x*x
         FMUL  s8, s1, s1 // y*y
         SUB   W1, W1, #1 // decrement pixel counter
         FMUL s12, s0, s1 // x*y
         FADD  s7, s3, s8 // x*x + y*y
         FSUB  s9, s3, s8 // x_new = x*x - y*y
         FADD s12,s12,s12 // 2*x*y
         FADD  s0, s9, s4 // x_new = x_new + a
         FCMP  s7,s15     // pixel 1 > 4 ? check if pixel 1 diverged
         FADD  s1,s12, s5 // y_new = y_new + b
         B.GT  end_pixel_3_diverged
         TBZ W1,31,iterate_pixel_3

          // maximum iterations reached
         ADD W4,W4,#max_iterations   // add to global iteration counter
         B check_end_of_line_pixel_4

         end_pixel_3_diverged:
         SUB W2,W12,W1          // correct counter
         ADD W4,W4, W2          // add to global iteration counter
         LSL W2,W2,#colour_shift// some colouring
         STR W2,[X25,W8,UXTW#2] // write pixel to buffer

      //############ End of Line Pixel 4 Handling  ############

      check_end_of_line_pixel_4:

      MOV v0.16B,v24.16B // restore state of v0 for other pixels 
      MOV v1.16B,v25.16B // restore state of v1 for other pixels
 
      TBNZ W11,#3,next_line //test for 0b1000
 
      INS v0.s[0],v0.s[3] 
      INS v1.s[0],v1.s[3] 
      INS v4.s[0],v4.s[3] 
      MOV W1,v2.s[3]  
      iterate_pixel_4:
         FMUL  s3, s0, s0 // x*x
         FMUL  s8, s1, s1 // y*y
         SUB   W1, W1, #1 // decrement pixel counter
         FMUL s12, s0, s1 // x*y
         FADD  s7, s3, s8 // x*x + y*y
         FSUB  s9, s3, s8 // x_new = x*x - y*y
         FADD s12,s12,s12 // 2*x*y
         FADD  s0, s9, s4 // x_new = x_new + a
         FCMP  s7,s15     // pixel 1 > 4 ? check if pixel 1 diverged
         FADD  s1,s12, s5 // y_new = y_new + b
         B.GT  end_pixel_4_diverged
         TBZ W1,31,iterate_pixel_4
 
         // maximum iterations reached
         ADD W4,W4,#max_iterations   // add to global iteration counter
         B next_line

         end_pixel_4_diverged:
         SUB W2,W12,W1          // correct counter
         ADD W4,W4, W2          // add to global iteration counter
         LSL W2,W2,#colour_shift// some colouring
         STR W2,[X25,W9,UXTW#2] // write pixel to buffer

   next_line:

// atomic update of global y_line counter
   try_again_update_global_y_line_counter:
     LDAXR W6,[X13]        //try LDXR or LDAXR
     ADD   W27,W6,#1
     STLXR W7,W27,[X13]    //try STXR or STLXR
   CBNZ W7,try_again_update_global_y_line_counter
   
   CMP W27,#fractal_size   //may be check can already be done within the atomic update, check performance
   B.LO y_loop             //use lower check due to overflow of other cores
   
// prepare to return to C routine
// restore x17...x30, d8...d15
   LDP x17,x18,[sp],#16
   LDP x19,x20,[sp],#16
   LDP x21,x22,[sp],#16
   LDP x23,x24,[sp],#16
   LDP x25,x26,[sp],#16
   LDP x27,x28,[sp],#16
   LDP x29,x30,[sp],#16
   LDP  d8, d9,[sp],#16
   LDP d10,d11,[sp],#16
   LDP d12,d13,[sp],#16
   LDP d14,d15,[sp],#16
   LDP x1, x2, [sp],#48  //get address for iter count storage

// atomic update of global iteration counter
   try_again_update_global_iteration_counter:
     LDAXR X9,[X1]         //try LDXR or LDAXR
     ADD   X6,X9,X4
     STLXR W7,X6,[X1]      //try STXR or STLXR
   CBNZ W7,try_again_update_global_iteration_counter
      
RET

.data
.align 4
frac_data:
.single -0.1450 //+ 0 ro
.single -0.1750 //+ 4 ru
.single -1.0200 //+ 8 io
.single   600.0 //+12 fractal_size
.single    0.0  //+16 
.single    1.0
.single    2.0
.single    3.0  
