//----------------------------------------------------------
// FracNEON_sp_v8a_mt2_to Assembler routine by Michael Kübel
//----------------------------------------------------------
// extern "C" uint64_t CalcMandelbrotASM_FracNEON_sp_v8a_mt2_to(const uint32_t* p,const uint64_t* iterations,const uint32_t* y,const uint32_t y_start);
//
//  v0 = x0123
//  v1 = y0123
//  v2 = calc p0123 (x*x),...
//  v3 = calc p0123 (y*y),...
//  v4 = calc p0123 (x*x),...
//  v5 = x*x+y*y save iter p0123 0
//  v6 = x*x+y*y save iter p0123 1
//  v7 = counter p0123
//  v8 = a0123
//
//  v9 = x4567
// v10 = y4567
// v11 = calc p4567 (x*x),...
// v12 = calc p4567 (y*y),...
// v13 = calc p4567 (x*x),...
// v14 = x*x+y*y save iter p4567 0
// v15 = x*x+y*y save iter p4567 1
// v16 = counter p4567
// v17 = a4567
//
// v27 = b01234567891011
//          [3]|   [2]|   [1]|   [0]
// v28 =    4.0|   4.0|   4.0|   4.0   diverge limit float
// v29 =      4|     4|     4|     4   counter add integer
// v30 =      -|    io|    ru| delta
// s31 = dummy for calculation
//
//  w6, w7, w8, w9 = plot x0 x1  x2  x3
// w10             = global plot x
// w15,w16,w17,w18 = plot x4 x5  x6, x7
// x23 = backup address for end of line
// w24 = pixel plot reg for end of line subroutines
// x25 = pixel line position address
// x26 = pixel array start address

fractal_size   =  600 // size of fractal in pixels
max_iterations = 4096 // maximum iterations
colour_shift   =    3 // shift colours left

.text

.global CalcMandelbrotASM_FracNEON_sp_v8a_mt2_to
CalcMandelbrotASM_FracNEON_sp_v8a_mt2_to:

// to allow use of all regs and call further subroutines,
// backup x17...x30, d8...d15 due to calling conventions
// (x17 for align reasons and x18 due to system variations)
 STP x17,x18,[sp,#-224]!
 STP x19,x20,[sp,#  16]
 STP x21,x22,[sp,#  32]
 STP x23,x24,[sp,#  48]
 STP x25,x26,[sp,#  64]
 STP x27,x28,[sp,#  80]
 STP x29,x30,[sp,#  96]
 STP  d8, d9,[sp,# 112]
 STP d10,d11,[sp,# 128]
 STP d12,d13,[sp,# 144]
 STP d14,d15,[sp,# 160]
 STP x1,x2,  [sp,# 176] // X1 global address for iteration counter
                        // X2 global address for y 

 MOV X26,X0             // use X26 for pixel_buffer start address as X0 is used elsewhere
 MOV X13,X2             // backup address of global iteration counter as X2 is used elsewhere  
 MOV X5, X3             // set first y in X5 as X3 is used elsewhere
 MOV X4,#0              // init local iteration counter
 LDR X14,=frac_data     // pointer to fractal data
      
   FMOV v28.4s,#4.0            // diverge limit = 4.0 for each pixel
   MOVI v29.4s,#2              // counter sub => 2 iterations per loop  
   MOV W12,#(max_iterations-1) // init maximum iterations - 1

   LDP S0,S1,[X14]       // ro (S0), ru (S20)
   LDR S2,[X14,#12]      // fractal size
   INS v30.s[1],v1.s[0]  // save ru in v30.s[1] for later ***
   FSUB S0,S0,S1
   FDIV S0,S0,S2         // delta_sp = (ro-ru)/fractal_size
   INS v30.s[0],v0.s[0]  // save delta in v30.s[0] for later ***

   y_loop:
      MOV W6,#fractal_size  // get pixel_buffer_width
      MUL W6,W6,W5          // pixel_buffer_width * y_counter 
      ADD X25,X26,W6,UXTW#2 // Line position = pixel_buffer + (pixel_buffer_width * y)<<2

      LDR S1,[X14,#8]       // io
      UCVTF S2,W5           // get y 
      INS v0.s[0],v30.s[0]  // get delta 
      FMUL S2,S2,S0         // y*delta         
      FSUB S1,S1,S2         // b0 = io - y * delta
      INS v30.s[2],v1.s[0]  // save io in v30.s[2] for later ***
      DUP v27.4s,v1.s[0]    // save io_save_sp = b0=b1=b2... in v27

      MOV  W6,#0            // plot pixel 0
      MOV  W7,#1            // plot pixel 1
      MOV  W8,#2            // plot pixel 2
      MOV  W9,#3            // plot pixel 3
      MOV W15,#4            // plot pixel 4
      MOV W16,#5            // plot pixel 5
      MOV W17,#6            // plot pixel 6
      MOV W18,#7            // plot pixel 7
      MOV W10,#7            // plot pixel global init 
  
      LDP  q1,q2,[X14,#16]      // v1 =  0.0      1.0       2.0       3.0
                                // v2 =  4.0      5.0       6.0       7.0
      DUP  v3.4s,v30.s[1]       // v3 =   ru       ru        ru        ru
      FMUL v0.4s,v1.4s,v30.s[0] // v0 =  d*x      d*(x+1)   d*(x+2)   d*(x+3)
      FADD v0.4s,v0.4s,v3.4s    // v0 = ru+d*x  ru+d*(x+1) ru+d*(x+2)  ru+d*(x+3)
      FMUL v9.4s,v2.4s,v30.s[0] // v9 =  d*(x+4)  d*(x+5)   d*(x+6)   d*(x+7)
      FADD v9.4s,v9.4s,v3.4s    // v9 = ru+d*x+4) ru+d*(x+5) ru+d*(x+6)  ru+d*(x+7)

      MOV  v8.16B,v0.16B        // save a0,a1,a2,a3
      MOV v17.16B,v9.16B        // save a4,a5,a6,a7

      MOV  v1.16B,v27.16B       // init y0123 = b0
      MOV v10.16B,v27.16B       // init y4567 = b0
      DUP  v7.4s,W12            // init iteration counters p0123 = max_iter-1
      DUP v16.4s,W12            // init iteration counters p4567 = max_iter-1

      MOV X11,#0                // init end of line flags for each pixel

      iteration_loop:
      iteration_loop_continue_0123:
         //1.Iteration pixel 0123
         FMUL  v2.4s, v0.4s, v0.4s   // x*x
         FMUL  v3.4s, v1.4s, v1.4s   // y*y
         FMUL  v4.4s, v0.4s, v1.4s   // x*y
         FADD  v5.4s, v2.4s, v3.4s   // x*x + y*y *** backup in v5
         FSUB  v0.4s, v2.4s, v3.4s   // x_new = x*x - y*y
         FADD  v1.4s, v4.4s, v4.4s   // y_new = 2*x*y
         FADD  v0.4s, v0.4s, v8.4s   // x_new = x_new + a0123
         FADD  v1.4s, v1.4s,v27.4s   // y_new = y_new + b
         //1.Iteration pixel 4567
         FMUL v11.4s, v9.4s, v9.4s   // x*x
         FMUL v12.4s,v10.4s,v10.4s   // y*y
         FMUL v13.4s, v9.4s,v10.4s   // x*y
         FADD v14.4s,v11.4s,v12.4s   // x*x + y*y *** backup in v14
         FSUB  v9.4s,v11.4s,v12.4s   // x_new = x*x - y*y
         FADD v10.4s,v13.4s,v13.4s   // y_new = 2*x*y
         FADD  v9.4s, v9.4s,v17.4s   // x_new = x_new + a4567
         FADD v10.4s,v10.4s,v27.4s   // y_new = y_new + b
         //2.Iteration pixel 0123
         FMUL  v2.4s, v0.4s, v0.4s   // x*x
         FMUL  v3.4s, v1.4s, v1.4s   // y*y
         FMUL  v4.4s, v0.4s, v1.4s   // x*y
         FADD  v6.4s, v2.4s, v3.4s   // x*x + y*y *** backup in v6
         FSUB  v0.4s, v2.4s, v3.4s   // x_new = x*x - y*y
         FADD  v1.4s, v4.4s, v4.4s   // y_new = 2*x*y
         FCMGT v2.4s, v6.4s,v28.4s   // > 4 ?
         FADD  v0.4s, v0.4s, v8.4s   // x_new = x_new + a0123
         FADD  v1.4s, v1.4s,v27.4s   // y_new = y_new + b
         SUB   v7.4s, v7.4s,v29.4s   // decrement all pixel counter p0123 by 2 
         UMAXV S3,v2.4s
         UMAXV S4,v7.4s
         MOV   W0,v3.s[0]
         MOV   W1,v4.s[0]
         TST W0,W0                 // faster than TBNZ here...check again later
         B.MI diverged_0123        // 1 or more pixel diverged
         TST W1,W1                 // faster than TBNZ here...check again later
         B.MI max_iterations_0123  // 1 or more pixel reached max iter
      iteration_loop_continue_4567:
         //2.Iteration pixel 4567
         FMUL v11.4s, v9.4s, v9.4s   // x*x
         FMUL v12.4s,v10.4s,v10.4s   // y*y
         FMUL v13.4s, v9.4s,v10.4s   // x*y
         FADD v15.4s,v11.4s,v12.4s   // x*x + y*y *** backup in v15
         FSUB  v9.4s,v11.4s,v12.4s   // x_new = x*x - y*y
         FADD v10.4s,v13.4s,v13.4s   // y_new = 2*x*y
         FCMGT v11.4s,v15.4s,v28.4s  // > 4 ?
         FADD  v9.4s, v9.4s,v17.4s   // x_new = x_new + a4567
         FADD v10.4s,v10.4s,v27.4s   // y_new = y_new + b
         SUB  v16.4s,v16.4s,v29.4s   // decrement all pixel counter p4567 by 2 
         UMAXV S12,v11.4s
         UMAXV S13,v16.4s
         MOV   W28,v12.s[0]
         MOV   W29,v13.s[0]
         TST W28,W28               // faster than TBNZ here...check again later
         B.MI diverged_4567        // 1 or more pixel diverged
         TST W29,W29               // faster than TBNZ here...check again later
         B.MI max_iterations_4567  // 1 or more pixel reached max iter
      B iteration_loop
 
      //############ Pixel 0 Diverge Handling ############
      diverged_0123:
      FCMP S6,S28                // > 4 ? pixel 0 diverged  v6.s[0]
      B.LE diverged_123          // not ? -> check pixel 1,2,3
	  
      UMOV  W0,v7.s[0]           // get iterations pixel 0  v7.s[0]
      FCMP  S5,S28               // > 4 ? Diverged at 1st ? v5.s[0]
      CSINC W0,W0,W0,LE          // if => inc counter

        SUB   W2,W12,W0          // correct counter
        ADD   W10,W10,#1         // inc x
        ADD   W4,W4, W2          // add to global iteration counter

        LSL   W2,W2,#colour_shift// some colouring
        STR   W2,[X25,W6,UXTW#2] // write pixel to buffer

        CMP   W10,#fractal_size  // end of line reached ?
        B.LT skip_endofline_p0
          ORR W11,W11,#0b00000001// set end of line flags for pixel 0
          B   diverged_123       // but still check if also another diverged
        skip_endofline_p0:

        INS   v7.s[0],W12        // transfer counter init to v7.s[0]
        MOV   W6,W10             // save new x in new plot x1
        UCVTF S31,W10
        INS   v2.s[0],v30.s[0]   // S2 = delta v2 not needed anymore 
        INS   v3.s[0],v30.s[1]   // S3 =    ru v3 not needed anymore 
        FMUL  S2,S2,S31          // new x0 = x0*d
        FADD  S2,S2,S3           // new x0 = x0*d + ru
        INS   v0.s[0], v2.s[0]   // place new x0 in correct vector v0.s[0]
        INS   v8.s[0], v2.s[0]   // new x0 also = new a0
        INS   v1.s[0],v30.s[2]   // io => get saved y3=y2=y1=y0
        //just continue checking next pixel

      //############ Pixel 1 Diverge Handling ############
      diverged_123:
      INS  v2.s[0],v6.s[1]       // get pixel 1 from v6.s[1]
      FCMP S2,S28                // > 4 ? pixel 1 diverged  v6.s[1]
      B.LE diverged_23           // not ? -> check pixel 3,4

      INS   v2.s[0],v5.s[1]      // get diverge value from 1st iter
      UMOV  W0,v7.s[1]           // get iterations pixel 1  v7.s[1]
      FCMP  S2,S28               // > 4 ? Diverged at 2nd ? v5.s[1]
      CSINC W0,W0,W0,LE          // if => inc counter

        SUB   W2,W12,W0          // correct counter
        ADD   W10,W10,#1         // inc x
        ADD   W4,W4, W2          // add to global iteration counter

        LSL   W2,W2,#colour_shift// some colouring
        STR   W2,[X25,W7,UXTW#2] // write pixel to buffer

        CMP   W10,#fractal_size  // end of line reached ?
        B.LT skip_endofline_p1
          ORR W11,W11,#0b00000010// set end of line flags for pixel 1
          B   diverged_23        // but still check if also another diverged
        skip_endofline_p1:

        INS   v7.s[1],W12        // transfer counter init to v7.s[1]
        MOV   W7,W10             // save new x in new plot x1
        UCVTF S31,W10
        INS   v2.s[0],v30.s[0]   // S2 = delta v2 not needed anymore 
        INS   v3.s[0],v30.s[1]   // S3 =    ru v3 not needed anymore 
        FMUL  S2,S2,S31          // new x1 = x1*d
        FADD  S2,S2,S3           // new x1 = ru + x1*d
        INS   v0.s[1], v2.s[0]   // place new x1 in correct vector v0.s[1]
        INS   v8.s[1], v2.s[0]   // new x1 also = new a1    
        INS   v1.s[1],v30.s[2]   // io => get saved y3=y2=y1=y0
        //just continue checking next pixel

      //############ Pixel 2 Diverge Handling ############
      diverged_23:
      INS  v2.s[0],v6.s[2]       // get pixel 2 from v6.4[2]
      FCMP S2,S28                // > 4 ? pixel 2 diverged
      B.LE diverged_3            // not ? -> check pixel 3  v6.s[1]

      INS   v2.s[0],v5.s[2]      // get diverge value from 1st ter
      UMOV  W0,v7.s[2]           // get iterations pixel 2  v7.s[2]
      FCMP  S2,S28               // > 4 ? Diverged at 1st ? v5.s[2]
      CSINC W0,W0,W0,LE          // if => inc counter
 
        SUB   W2,W12,W0          // correct counter
        ADD   W10,W10,#1         // inc x
        ADD   W4,W4, W2          // add to global iteration counter

        LSL   W2,W2,#colour_shift// some colouring
        STR   W2,[X25,W8,UXTW#2] // write pixel to buffer

        CMP   W10,#fractal_size  // end of line reached ?
        B.LT skip_endofline_p2
          ORR W11,W11,#0b00000100// set end of line flags for pixel 2
          B   diverged_3         // but still check if also another diverged
        skip_endofline_p2:

        INS   v7.s[2],W12        // transfer counter init to v2.s[7]
        MOV   W8,W10             // save new x in new plot_x3
        UCVTF S31,W10
        INS   v2.s[0],v30.s[0]   // S2 = delta v2 not needed anymore 
        INS   v3.s[0],v30.s[1]   // S3 =    ru v3 not needed anymore 
        FMUL  S2,S2,S31          // new x2 = x2*d
        FADD  S2,S2,S3           // new x2 = ru + x2*d
        INS   v0.s[2], v2.s[0]   // place new x2 in correct vector v0.s[2]
        INS   v8.s[2], v2.s[0]   // new x2 also = new a2    
        INS   v1.s[2],v30.s[2]   // io => get saved y3=y2=y1=y0
        //just continue checking next pixel

      //############ Pixel 3 Diverge Handling ############
      diverged_3:
      INS  v2.s[0],v6.s[3]       // get pixel 3 from v6.4[3]
      FCMP S2,S28                // > 4 ? pixel 3 diverged
      B.LE end_div_check_0123    // not ? -> continue

      INS   v2.s[0],v5.s[3]      // get diverge value from 1st iter
      UMOV  W0,v7.s[3]           // get iterations pixel 3   v7.4s[3]
      FCMP  S2,S28               // > 4 ? Diverged at 1st ?  v5.4s[3]
      CSINC W0,W0,W0,LE          // if => inc counter

        SUB   W2,W12,W0          // correct counter
        ADD   W10,W10,#1         // inc x
        ADD   W4,W4, W2          // add to global iteration counter

        LSL   W2,W2,#colour_shift// some colouring
        STR   W2,[X25,W9,UXTW#2] // write pixel to buffer

        CMP   W10,#fractal_size  // end of line reached ?
        B.LT skip_endofline_p3
          ORR W11,W11,#0b00001000// set end of line flags for pixel 3
          B   end_div_check_0123  // but still check if also another diverged
        skip_endofline_p3:

        INS   v7.s[3],W12        // transfer counter init to v2.s[3]
        MOV   W9,W10             // save new x in new plot_x3
        UCVTF S31,W10
        INS   v2.s[0],v30.s[0]   // S2 = delta v2 not needed anymore 
        INS   v3.s[0],v30.s[1]   // S3 =    ru v3 not needed anymore 
        FMUL  S2,S2,S31          // new x3 = x3*d
        FADD  S2,S2,S3           // new x3 = ru + x3*d
        INS   v0.s[3], v2.s[0]   // place new x3 in correct vector v0.s[3]
        INS   v8.s[3], v2.s[0]   // new x3 also = new a3
        INS   v1.s[3],v30.s[2]   // io => get saved y3=y2=y1=y0

      end_div_check_0123:

    UMAXV S31,v7.4s                      // check if max iter p0123
    MOV   W0,v31.s[0]
    TBNZ  W0,31,max_iterations_0123      // 1 or more pixel reached also max iter
    CBZ W11,iteration_loop_continue_4567 // if not, was end of line reached ?
    B end_of_line

      //############ Pixel 0 Max Iterations Handling ############
      max_iterations_0123: 
      MOV W0,v7.s[0]               // check if pixel 0 reach max iter
      TBZ  W0,31,max_iterations_123 // not ? -> check pixel 1,2,3

      ADD W10,W10,#1           // inc x
      ADD W4,W4,#max_iterations// add to global iteration counter
      CMP W10,#fractal_size    // end of line reached ?
      B.LT max_skip_endofline_p0
        ORR W11,W11,#0b00000001// set end of line flags for pixel 0
        B   max_iterations_123 // but still check if also another reached max iter
      max_skip_endofline_p0:

      INS   v7.s[0],W12        // transfer counter init to v7.s[0]
      MOV   W6,W10             // save new x in new plot x1
      UCVTF S31,W10
      INS   v2.s[0],v30.s[0]   // S2 = delta v2 not needed anymore 
      INS   v3.s[0],v30.s[1]   // S3 =    ru v3 not needed anymore 
      FMUL  S2,S2,S31          // new x0 = x0*d
      FADD  S2,S2,S3           // new x0 = x0*d + ru
      INS   v0.s[0], v2.s[0]   // place new x0 in correct vector v0.s[0]
      INS   v8.s[0], v2.s[0]   // new x0 also = new a0
      INS   v1.s[0],v30.s[2]   // io => get saved y3=y2=y1=y0
      //just continue checking next pixel

      //############ Pixel 1 Max Iterations Handling ############
      max_iterations_123:
      MOV W0,v7.s[1]              // check if pixel 1 reach max iter
      TBZ W0,31,max_iterations_23 // not ? -> check pixel 2,3

      ADD W10,W10,#1           // inc x
      ADD W4,W4,#max_iterations// add to global iteration counter
      CMP W10,#fractal_size    // end of line reached ?
      B.LT max_skip_endofline_p1
        ORR W11,W11,#0b00000010// set end of line flags for pixel 1
        B   max_iterations_23  // but still check if also another reached max iter
      max_skip_endofline_p1:

      INS   v7.s[1],W12        // transfer counter init to v7.s[1]
      MOV   W7,W10             // save new x in new plot x1
      UCVTF S31,W10
      INS   v2.s[0],v30.s[0]   // S2 = delta v2 not needed anymore 
      INS   v3.s[0],v30.s[1]   // S3 =    ru v3 not needed anymore 
      FMUL  S2,S2,S31          // new x1 = x1*d
      FADD  S2,S2,S3           // new x1 = ru + x1*d
      INS   v0.s[1], v2.s[0]   // place new x1 in correct vector v0.s[1]
      INS   v8.s[1], v2.s[0]   // new x1 also = new a1    
      INS   v1.s[1],v30.s[2]   // io => get saved y3=y2=y1=y0
      //just continue checking next pixel

      //############ Pixel 2 Max Iterations Handling ############
      max_iterations_23:
      MOV W0,v7.s[2]             // check if pixel 2 reach max iter
      TBZ  W0,31,max_iterations_3 // not ? -> check pixel 3

      ADD W10,W10,#1           // inc x
      ADD W4,W4,#max_iterations// add to global iteration counter
      CMP W10,#fractal_size    // end of line reached ?
      B.LT max_skip_endofline_p2
        ORR W11,W11,#0b00000100// set end of line flags for pixel 2
        B   max_iterations_3   // but still check if also another reached max iter
      max_skip_endofline_p2:

      INS   v7.s[2],W12        // transfer counter init to v2.s[7]
      MOV   W8,W10             // save new x in new plot_x3
      UCVTF S31,W10
      INS   v2.s[0],v30.s[0]   // S2 = delta v2 not needed anymore 
      INS   v3.s[0],v30.s[1]   // S3 =    ru v3 not needed anymore 
      FMUL  S2,S2,S31          // new x2 = x2*d
      FADD  S2,S2,S3           // new x2 = ru + x2*d
      INS   v0.s[2], v2.s[0]   // place new x2 in correct vector v0.s[2]
      INS   v8.s[2], v2.s[0]   // new x2 also = new a2    
      INS   v1.s[2],v30.s[2]   // io => get saved y3=y2=y1=y0
      //just continue checking next pixel

      //############ Pixel 3 Max Iterations Handling ############
      max_iterations_3:
      MOV W0,v7.s[3]              // check if pixel 3 reach max iter
      TBZ  W0,31,end_max_iter_0123 // not ? -> end max iter check

      ADD W10,W10,#1           // inc x
      ADD W4,W4,#max_iterations// add to global iteration counter
      CMP W10,#fractal_size    // end of line reached ?
      B.LT max_skip_endofline_p3
        ORR W11,W11,#0b00001000// set end of line flags for pixel 4
        B   end_max_iter_0123  // now all are checked for everything
      max_skip_endofline_p3:

      INS   v7.s[3],W12        // transfer counter init to v2.s[3]
      MOV   W9,W10             // save new x in new plot_x3
      UCVTF S31,W10
      INS   v2.s[0],v30.s[0]   // S2 = delta v2 not needed anymore 
      INS   v3.s[0],v30.s[1]   // S3 =    ru v3 not needed anymore 
      FMUL  S2,S2,S31          // new x3 = x3*d
      FADD  S2,S2,S3           // new x3 = ru + x3*d
      INS   v0.s[3], v2.s[0]   // place new x3 in correct vector v0.s[3]
      INS   v8.s[3], v2.s[0]   // new x3 also = new a3
      INS   v1.s[3],v30.s[2]   // io => get saved y3=y2=y1=y0

   end_max_iter_0123:
   CBZ W11,iteration_loop_continue_4567
   B end_of_line

      //############ Pixel 4 Diverge Handling ############
      diverged_4567:
      FCMP S15,S28               // > 4 ? pixel 4 diverged  v15.s[0]
      B.LE diverged_567          // not ? -> check pixel 5,6,7
	  
      UMOV  W0,v16.s[0]          // get iterations pixel 4  v16.s[0]
      FCMP  S14,S28              // > 4 ? Diverged at 1st ? v14.s[0]
      CSINC W0,W0,W0,LE          // if => inc counter

        SUB   W2,W12,W0          // correct counter
        ADD   W10,W10,#1         // inc x
        ADD   W4,W4, W2          // add to global iteration counter

        LSL   W2,W2,#colour_shift// some colouring
        STR   W2,[X25,W15,UXTW#2]// write pixel to buffer

        CMP   W10,#fractal_size  // end of line reached ?
        B.LT skip_endofline_p4
          ORR W11,W11,#0b00010000// set end of line flags for pixel 4
          B   diverged_567       // but still check if also another diverged
        skip_endofline_p4:

        INS   v16.s[0],W12       // transfer counter init to v16.s[0]
        MOV   W15,W10            // save new x in new plot x4
        UCVTF S31,W10
        INS   v11.s[0],v30.s[0]  // S11 = delta v11 not needed anymore 
        INS   v12.s[0],v30.s[1]  // S12 =    ru v12 not needed anymore 
        FMUL  S11,S11,S31        // new x4 = x4*d
        FADD  S11,S11,S12        // new x4 = x4*d + ru
        INS    v9.s[0],v11.s[0]  // place new x4 in correct vector v9.s[0]
        INS   v17.s[0],v11.s[0]  // new x4 also = new a4
        INS   v10.s[0],v30.s[2]  // io => get saved y7=y6=y5=y4

      //############ Pixel 5 Diverge Handling ############
      diverged_567:
      INS  v11.s[0],v15.s[1]     // get pixel 5 from v15.s[1]
      FCMP S11,S28               // > 4 ? pixel 5 diverged  v15.s[1]
      B.LE diverged_67           // not ? -> check pixel 6,7

      INS   v11.s[0],v14.s[1]    // get diverge value from 1st iter
      UMOV  W0,v16.s[1]          // get iterations pixel 5  v16.s[1]
      FCMP  S11,S28              // > 4 ? Diverged at 1st ? v14.s[1]
      CSINC W0,W0,W0,LE          // if => inc counter

        SUB   W2,W12,W0          // correct counter
        ADD   W10,W10,#1         // inc x
        ADD   W4,W4, W2          // add to global iteration counter

        LSL   W2,W2,#colour_shift// some colouring
        STR   W2,[X25,W16,UXTW#2]// write pixel to buffer

        CMP   W10,#fractal_size  // end of line reached ?
        B.LT skip_endofline_p5
          ORR W11,W11,#0b00100000// set end of line flags for pixel 1
          B   diverged_67        // but still check if also another diverged
        skip_endofline_p5:

        INS   v16.s[1],W12       // transfer counter init to v16.s[1]
        MOV   W16,W10            // save new x in new plot x5
        UCVTF S31,W10
        INS   v11.s[0],v30.s[0]  // S11 = delta v11 not needed anymore 
        INS   v12.s[0],v30.s[1]  // S12 =    ru v12 not needed anymore 
        FMUL  S11,S11,S31        // new x5 = x5*d
        FADD  S11,S11,S12        // new x5 = x5*d + ru
        INS    v9.s[1],v11.s[0]  // place new x5 in correct vector v9.s[1]
        INS   v17.s[1],v11.s[0]  // new x5 also = new a5
        INS   v10.s[1],v30.s[2]  // io => get saved y7=y6=y5=y4
        //just continue checking next pixel

      //############ Pixel 6 Diverge Handling ############
      diverged_67:
      INS  v11.s[0],v15.s[2]     // get pixel 6 from v15.4[2]
      FCMP S11,S28               // > 4 ? pixel 6 diverged
      B.LE diverged_7            // not ? -> check pixel 6  v15.s[2]

      INS   v11.s[0],v14.s[2]    // get diverge value from 1st ter
      UMOV  W0,v16.s[2]          // get iterations pixel 6  v16.s[2]
      FCMP  S11,S28              // > 4 ? Diverged at 1st ? v14.s[2]
      CSINC W0,W0,W0,LE          // if => inc counter
 
        SUB   W2,W12,W0          // correct counter
        ADD   W10,W10,#1         // inc x
        ADD   W4,W4, W2          // add to global iteration counter

        LSL   W2,W2,#colour_shift// some colouring
        STR   W2,[X25,W17,UXTW#2]// write pixel to buffer

        CMP   W10,#fractal_size  // end of line reached ?
        B.LT skip_endofline_p6
          ORR W11,W11,#0b01000000// set end of line flags for pixel 2
          B   diverged_7         // but still check if also another diverged
        skip_endofline_p6:

        INS   v16.s[2],W12       // transfer counter init to v16.s[2]
        MOV   W17,W10            // save new x in new plot x6
        UCVTF S31,W10
        INS   v11.s[0],v30.s[0]  // S11 = delta v11 not needed anymore 
        INS   v12.s[0],v30.s[1]  // S12 =    ru v12 not needed anymore 
        FMUL  S11,S11,S31        // new x6 = x6*d
        FADD  S11,S11,S12        // new x6 = x6*d + ru
        INS    v9.s[2],v11.s[0]  // place new x6 in correct vector v9.s[2]
        INS   v17.s[2],v11.s[0]  // new x6 also = new a6
        INS   v10.s[2],v30.s[2]  // io => get saved y7=y6=y5=y4
        //just continue checking next pixel

      //############ Pixel 7 Diverge Handling ############
      diverged_7:
      INS  v11.s[0],v15.s[3]     // get pixel 7 from v15.4[3]
      FCMP S11,S28               // > 4 ? pixel 7 diverged
      B.LE end_div_check_4567    // not ? -> continue

      INS   v11.s[0],v14.s[3]    // get diverge value from 1st iter
      UMOV  W0,v16.s[3]          // get iterations pixel 7   v16.4s[3]
      FCMP  S11,S28              // > 4 ? Diverged at 1st ?  v14.4s[3]
      CSINC W0,W0,W0,LE          // if => inc counter

        SUB   W2,W12,W0          // correct counter
        ADD   W10,W10,#1         // inc x
        ADD   W4,W4, W2          // add to global iteration counter

        LSL   W2,W2,#colour_shift// some colouring
        STR   W2,[X25,W18,UXTW#2]// write pixel to buffer

        CMP   W10,#fractal_size  // end of line reached ?
        B.LT skip_endofline_p7
          ORR W11,W11,#0b10000000// set end of line flags for pixel 3
          B   end_div_check_4567 // but still check if also another diverged
        skip_endofline_p7:

        INS   v16.s[3],W12       // transfer counter init to v16.s[3]
        MOV   W18,W10            // save new x in new plot_x3
        UCVTF S31,W10
        INS   v11.s[0],v30.s[0]  // S11 = delta v11 not needed anymore 
        INS   v12.s[0],v30.s[1]  // S12 =    ru v12 not needed anymore 
        FMUL  S11,S11,S31        // new x7 = x7*d
        FADD  S11,S11,S12        // new x7 = x7*d + ru
        INS    v9.s[3],v11.s[0]  // place new x7 in correct vector v9.s[3]
        INS   v17.s[3],v11.s[0]  // new x7 also = new a7
        INS   v10.s[3],v30.s[2]  // io => get saved y7=y6=y5=y4

      end_div_check_4567:

    UMAXV S31,v16.4s                     // check if max iter p4567
    MOV   W0,v31.s[0]
    TBNZ  W0,31,max_iterations_4567      // 1 or more pixel reached also max iter
    CBZ W11,iteration_loop_continue_0123 // if not, was end of line reached ?
    B end_of_line

      //############ Pixel 4 Max Iterations Handling ############
      max_iterations_4567: 
      UMOV W0,v16.s[0]              // check if pixel 4 reach max iter
      TBZ  W0,31,max_iterations_567 // not ? -> check pixel 5,6,7

      ADD W10,W10,#1           // inc x
      ADD W4,W4,#max_iterations// add to global iteration counter
      CMP W10,#fractal_size    // end of line reached ?
      B.LT max_skip_endofline_p4
        ORR W11,W11,#0b00010000// set end of line flags for pixel 4
        B   max_iterations_567 // but still check if also another reached max iter
      max_skip_endofline_p4:

      INS   v16.s[0],W12       // transfer counter init to v16.s[0]
      MOV   W15,W10            // save new x in new plot x4
      UCVTF S31,W10
      INS   v11.s[0],v30.s[0]  // S11 = delta v11 not needed anymore 
      INS   v12.s[0],v30.s[1]  // S12 =    ru v12 not needed anymore 
      FMUL  S11,S11,S31        // new x4 = x4*d
      FADD  S11,S11,S12        // new x4 = x4*d + ru
      INS    v9.s[0],v11.s[0]  // place new x4 in correct vector v9.s[0]
      INS   v17.s[0],v11.s[0]  // new x4 also = new a4
      INS   v10.s[0],v30.s[2]  // io => get saved y7=y6=y5=y4

      //############ Pixel 5 Max Iterations Handling ############
      max_iterations_567:
      UMOV W0,v16.s[1]             // check if pixel 5 reach max iter
      TBZ  W0,31,max_iterations_67 // not ? -> check pixel 6,7

      ADD W10,W10,#1           // inc x
      ADD W4,W4,#max_iterations// add to global iteration counter
      CMP W10,#fractal_size    // end of line reached ?
      B.LT max_skip_endofline_p5
        ORR W11,W11,#0b00100000// set end of line flags for pixel 5
        B   max_iterations_67  // but still check if also another reached max iter
      max_skip_endofline_p5:

      INS   v16.s[1],W12       // transfer counter init to v16.s[1]
      MOV   W16,W10            // save new x in new plot x5
      UCVTF S31,W10
      INS   v11.s[0],v30.s[0]  // S11 = delta v11 not needed anymore 
      INS   v12.s[0],v30.s[1]  // S12 =    ru v12 not needed anymore 
      FMUL  S11,S11,S31        // new x5 = x5*d
      FADD  S11,S11,S12        // new x5 = x5*d + ru
      INS    v9.s[1],v11.s[0]  // place new x5 in correct vector v9.s[1]
      INS   v17.s[1],v11.s[0]  // new x5 also = new a5
      INS   v10.s[1],v30.s[2]  // io => get saved y7=y6=y5=y4
      //just continue checking next pixel

      //############ Pixel 6 Max Iterations Handling ############
      max_iterations_67:
      UMOV W0,v16.s[2]            // check if pixel 6 reach max iter
      TBZ  W0,31,max_iterations_7 // not ? -> check pixel 7

      ADD W10,W10,#1           // inc x
      ADD W4,W4,#max_iterations// add to global iteration counter
      CMP W10,#fractal_size    // end of line reached ?
      B.LT max_skip_endofline_p6
        ORR W11,W11,#0b01000000// set end of line flags for pixel 6
        B   max_iterations_7   // but still check if also another reached max iter
      max_skip_endofline_p6:

      INS   v16.s[2],W12       // transfer counter init to v16.s[2]
      MOV   W17,W10            // save new x in new plot x6
      UCVTF S31,W10
      INS   v11.s[0],v30.s[0]  // S11 = delta v11 not needed anymore 
      INS   v12.s[0],v30.s[1]  // S12 =    ru v12 not needed anymore 
      FMUL  S11,S11,S31        // new x6 = x6*d
      FADD  S11,S11,S12        // new x6 = x6*d + ru
      INS    v9.s[2],v11.s[0]  // place new x6 in correct vector v9.s[2]
      INS   v17.s[2],v11.s[0]  // new x6 also = new a6
      INS   v10.s[2],v30.s[2]  // io => get saved y7=y6=y5=y4
      //just continue checking next pixel

      //############ Pixel 7 Max Iterations Handling ############
      max_iterations_7:
      UMOV W0,v16.s[3]             // check if pixel 7 reach max iter
      TBZ  W0,31,end_max_iter_4567 // not ? -> end max iter check

      ADD W10,W10,#1           // inc x
      ADD W4,W4,#max_iterations// add to global iteration counter
      CMP W10,#fractal_size    // end of line reached ?
      B.LT max_skip_endofline_p7
        ORR W11,W11,#0b10000000// set end of line flags for pixel 7
        B   end_max_iter_4567  // now all are checked for everything
      max_skip_endofline_p7:

      INS   v16.s[3],W12       // transfer counter init to v16.s[3]
      MOV   W18,W10            // save new x in new plot_x3
      UCVTF S31,W10
      INS   v11.s[0],v30.s[0]  // S11 = delta v11 not needed anymore 
      INS   v12.s[0],v30.s[1]  // S12 =    ru v12 not needed anymore 
      FMUL  S11,S11,S31        // new x7 = x7*d
      FADD  S11,S11,S12        // new x7 = x7*d + ru
      INS    v9.s[3],v11.s[0]  // place new x7 in correct vector v9.s[3]
      INS   v17.s[3],v11.s[0]  // new x7 also = new a7
      INS   v10.s[3],v30.s[2]  // io => get saved y7=y6=y5=y4

      end_max_iter_4567:
      CBZ W11,iteration_loop_continue_0123

      //############ End of Line Handling  ############
//  v0 = x0123
//  v1 = y0123
//  v2 = calc p0123 (x*x),...
//  v3 = calc p0123 (y*y),...
//  v4 = calc p0123 (x*x),...
//  v5 = x*x+y*y save iter p0123 0
//  v6 = x*x+y*y save iter p0123 1
//  v7 = counter p0123
//  v8 = a0123

//  v9 = x4567
// v10 = y4567
// v11 = calc p4567 (x*x),...
// v12 = calc p4567 (y*y),...
// v13 = calc p4567 (x*x),...
// v14 = x*x+y*y save iter p4567 0
// v15 = x*x+y*y save iter p4567 1
// v16 = counter p4567
// v17 = a4567

      end_of_line:

      //############ End of Line Pixel 0 Handling  ############
      check_end_of_line_pixel_0:
      TBNZ W11,#0,check_end_of_line_pixel_1
        STP Q0,Q1,[sp,#192]  //backup reg v0,v1 (x0123,y0123)
        INS v31.s[0],v8.s[0] //get a0     
        MOV       W0,v7.s[0] //get iterations p0
        MOV W24,W6           //get plot_x
        BL end_of_line_0123_iter
        LDP Q0,Q1,[sp,#192]  //restore reg v0,v1 (x0123,y0123)

      //############ End of Line Pixel 1 Handling  ############
      check_end_of_line_pixel_1:
      TBNZ W11,#1,check_end_of_line_pixel_2
        STP Q0,Q1,[sp,#192]  //backup reg v0,v1 (x0123,y0123)
        INS  v0.s[0],v0.s[1] //get x1
        INS  v1.s[0],v1.s[1] //get y1
        INS v31.s[0],v8.s[1] //get a1
        MOV       W0,v7.s[1] //get iterations p1
        MOV W24,W7           //get plot_x
        BL end_of_line_0123_iter
        LDP Q0,Q1,[sp,#192]  //restore reg v0,v1 (x0123,y0123)

      //############ End of Line Pixel 2 Handling  ############
      check_end_of_line_pixel_2:
      TBNZ W11,#2,check_end_of_line_pixel_3
        STP Q0,Q1,[sp,#192]  //backup reg v0,v1 (x0123,y0123)
        INS  v0.s[0],v0.s[2] //get x2
        INS  v1.s[0],v1.s[2] //get y2
        INS v31.s[0],v8.s[2] //get a2
        MOV       W0,v7.s[2] //get iterations p2
        MOV W24,W8           //get plot_x
        BL end_of_line_0123_iter
        LDP Q0,Q1,[sp,#192]  //restore reg v0,v1 (x0123,y0123)

      //############ End of Line Pixel 3 Handling  ############
      check_end_of_line_pixel_3:
      TBNZ W11,#3,check_end_of_line_pixel_4
        STP Q0,Q1,[sp,#192]  //backup reg v0,v1 (x0123,y0123)
        INS  v0.s[0],v0.s[3] //get x3
        INS  v1.s[0],v1.s[3] //get y3
        INS v31.s[0],v8.s[3] //get a3
        MOV       W0,v7.s[3] //get iterations p3
        MOV W24,W9           //get plot_x
        BL end_of_line_0123_iter
        LDP Q0,Q1,[sp,#192]  //restore reg v0,v1 (x0123,y0123)

      //############ End of Line Pixel 4 Handling  ############
      check_end_of_line_pixel_4:
      TBNZ W11,#4,check_end_of_line_pixel_5
        INS  v0.s[0], v9.s[0] //get x4
        INS  v1.s[0],v10.s[0] //get y4
        INS v31.s[0],v17.s[0] //get a4
        INS  v5.s[0],v14.s[0] //get diverge backup
        MOV       W0,v16.s[0] //get iterations p4
        MOV W24,W15           //get plot_x
        BL end_of_line_4567_iter

      //############ End of Line Pixel 5 Handling  ############
      check_end_of_line_pixel_5:
      TBNZ W11,#5,check_end_of_line_pixel_6
        INS  v0.s[0], v9.s[1] //get x5
        INS  v1.s[0],v10.s[1] //get y5
        INS v31.s[0],v17.s[1] //get a5
        INS  v5.s[0],v14.s[1] //get diverge backup
        MOV       W0,v16.s[1] //get iterations p5
        MOV W24,W16           //get plot_x
        BL end_of_line_4567_iter

      //############ End of Line Pixel 6 Handling  ############
      check_end_of_line_pixel_6:
      TBNZ W11,#6,check_end_of_line_pixel_7
        INS  v0.s[0], v9.s[2] //get x6
        INS  v1.s[0],v10.s[2] //get y6
        INS v31.s[0],v17.s[2] //get a6
        INS  v5.s[0],v14.s[2] //get diverge backup
        MOV       W0,v16.s[2] //get iterations p6
        MOV W24,W17           //get plot_x
        BL end_of_line_4567_iter

      //############ End of Line Pixel 7 Handling  ############
      check_end_of_line_pixel_7:
      TBNZ W11,#7,next_line
        INS  v0.s[0], v9.s[3] //get x7
        INS  v1.s[0],v10.s[3] //get y7
        INS v31.s[0],v17.s[3] //get a7
        INS  v5.s[0],v14.s[3] //get diverge backup
        MOV       W0,v16.s[3] //get iterations p7
        MOV W24,W18           //get plot_x
        BL end_of_line_4567_iter

   next_line:

// atomic update of global y_line counter
   try_again_update_global_y_line_counter:
     LDAXR W6,[X13]        //try LDXR or LDAXR
     ADD   W5,W6,#1
     STLXR W7,W5,[X13]     //try STXR or STLXR
   CBNZ W7,try_again_update_global_y_line_counter
   
   CMP W5,#fractal_size    //may be check can already be done within the atomic update, check performance
   B.LO y_loop             //use lower check due to overflow of other cores
   
// prepare to return to C routine
// restore x17...x30, d8...d15
   LDP x17,x18,[sp],#16
   LDP x19,x20,[sp],#16
   LDP x21,x22,[sp],#16
   LDP x23,x24,[sp],#16
   LDP x25,x26,[sp],#16
   LDP x27,x28,[sp],#16
   LDP x29,x30,[sp],#16
   LDP  d8, d9,[sp],#16
   LDP d10,d11,[sp],#16
   LDP d12,d13,[sp],#16
   LDP d14,d15,[sp],#16
   LDP x1, x2, [sp],#48  //get address for iter count storage

// atomic update of global iteration counter
   try_again_update_global_iteration_counter:
     LDAXR X9,[X1]         //try LDXR or LDAXR
     ADD   X6,X9,X4
     STLXR W7,X6,[X1]      //try STXR or STLXR
   CBNZ W7,try_again_update_global_iteration_counter
      
RET

 //subroutines
   end_of_line_0123_iter:
      iterate_pixel_0123:
         FMUL  S2, S0, S0 // x*x
         FMUL  S3, S1, S1 // y*y
         FMUL  S4, S0, S1 // x*y
         SUB   W0, W0, #1 // decrement iteration counter
         FADD  S6, S2, S3 // x*x + y*y *** backup in v6
         FSUB  S0, S2, S3 // x_new = x*x - y*y
         FADD  S1, S4, S4 // y_new = 2*x*y
         FCMP  S6,S28     // > 4 ?
         FADD  S0, S0,S31 // x_new = x_new + a0123
         FADD  S1, S1,S27 // y_new = y_new + b
         B.GT  end_pixel_0123_diverged
       TBZ W0,31,iterate_pixel_0123
       ADD W4,W4,#max_iterations // add to global iteration counter
     RET
       end_pixel_0123_diverged:
       SUB W2,W12,W0             // correct counter
       ADD W4,W4, W2             // add to global iteration counter
       LSL W2,W2,#colour_shift   // some colouring
       STR W2,[X25,W24,UXTW#2]   // write pixel to buffer
     RET

   end_of_line_4567_iter:
      TST W11,#0b11110000        // if exit from p0123 then one iteration already done !
      B.EQ at_first_one_iter_already_done_4567
      iterate_pixel_4567:
         FMUL  S2, S0, S0 // x*x
         FMUL  S3, S1, S1 // y*y
         FMUL  S4, S0, S1 // x*y
         FADD  S5, S2, S3 // x*x + y*y
         FSUB  S0, S2, S3 // x_new = x*x - y*y
         FADD  S1, S4, S4 // y_new = 2*x*y
         FADD  S0, S0,S31 // x_new = x_new + a4567
         FADD  S1, S1,S27 // y_new = y_new + b
         at_first_one_iter_already_done_4567:
         SUB   W0, W0, #2 // decrement iteration counter
         FMUL  S2, S0, S0 // x*x
         FMUL  S3, S1, S1 // y*y
         FMUL  S4, S0, S1 // x*y
         FADD  S6, S2, S3 // x*x + y*y
         FSUB  S0, S2, S3 // x_new = x*x - y*y
         FADD  S1, S4, S4 // y_new = 2*x*y
         FCMP  S6,S28     // > 4 ?
         FADD  S0, S0,S31 // x_new = x_new + a4567
         FADD  S1, S1,S27 // y_new = y_new + b
         B.GT  end_pixel_4567_diverged
       TBZ W0,31,iterate_pixel_4567
       ADD W4,W4,#max_iterations // add to global iteration counter
     RET
       end_pixel_4567_diverged:
       FCMP  S5,S28      // > 4 ? Diverged at 1st ?
       CSINC W0,W0,W0,LE // if => inc counter
       SUB   W2,W12,W0   // correct counter
       ADD   W4,W4, W2   // add to global iteration counter
       LSL W2,W2,#colour_shift   // some colouring
       STR W2,[X25,W24,UXTW#2]   // write pixel to buffer
     RET

.data
.align 4
frac_data:
.single -0.1450 //+ 0 ro
.single -0.1750 //+ 4 ru
.single -1.0200 //+ 8 io
.single   600.0 //+12 fractal_size
.single    0.0  //+16
.single    1.0
.single    2.0
.single    3.0
.single    4.0
.single    5.0
.single    6.0
.single    7.0
.single    8.0
.single    9.0
.single   10.0
.single   11.0
